# Copyright (C) 2021  UAVCAN Consortium  <uavcan.org>
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import sys
import abc
from typing import Optional, Union
import dataclasses
import pyuavcan
from uavcan.register import Value_1 as Value

if sys.version_info >= (3, 9):
    from collections.abc import MutableMapping
else:  # pragma: no cover
    from typing import MutableMapping  # pylint: disable=ungrouped-imports


__all__ = ["Value", "Backend", "Entry", "BackendError"]


class BackendError(RuntimeError):
    """
    Unsuccessful storage transaction. This is a very low-level error representing a system configuration issue.
    """


@dataclasses.dataclass(frozen=True)
class Entry:
    value: Value
    mutable: bool


class Backend(MutableMapping[str, Entry]):
    """
    Register backend interface implementing the :class:`MutableMapping` interface.
    The registers are ordered lexicographically by name.
    """

    @property
    @abc.abstractmethod
    def location(self) -> str:
        """
        The physical storage location for the data (e.g., file name).
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def persistent(self) -> bool:
        """
        An in-memory DB is reported as non-persistent.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def close(self) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def index(self, index: int) -> Optional[str]:
        """
        :returns: Name of the register at the specified index or None if the index is out of range.
            See ordering requirements in the class docs.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def __setitem__(self, key: str, value: Union[Entry, Value]) -> None:
        """
        If the register does not exist, it is either created or nothing is done, depending on the implementation.
        If exists, it will be overwritten unconditionally with the specified value.
        Observe that the method accepts either :class:`Entry` or :class:`Value`.

        The value shall be of the same type as the register, the caller is responsible to ensure that
        (implementations may lift this restriction if the type can be changed).

        The mutability flag is ignored (it is intended mostly for the UAVCAN Register Interface, not for local use).
        """
        raise NotImplementedError

    def __repr__(self) -> str:
        return pyuavcan.util.repr_attributes(self, repr(self.location), persistent=self.persistent)



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import typing


def refragment(input_fragments: typing.Iterable[memoryview], output_fragment_size: int) -> typing.Iterable[memoryview]:
    """
    Repackages the data from the arbitrarily-sized input fragments into fixed-size output fragments while minimizing
    the amount of data copying. The last fragment is allowed to be smaller than the requested size.
    If the input iterable contains no fragments or all of them are empty, nothing will be yielded.

    This function is designed for use in transfer emission logic where it's often needed to split a large
    payload into several frames while avoiding unnecessary copying. The best case scenario is when the size
    of input blocks is a multiple of the output fragment size -- in this case no copy will be done.

    >>> list(map(bytes, refragment([memoryview(b'0123456789'), memoryview(b'abcdef')], 7)))
    [b'0123456', b'789abcd', b'ef']

    The above example shows a marginally suboptimal case where one copy is required:

    - ``b'0123456789'[0:7]``  --> output    ``b'0123456'``       (slicing, no copy)
    - ``b'0123456789'[7:10]`` --> temporary ``b'789'``           (slicing, no copy)
    - ``b'abcdef'[0:4]``      --> output    ``b'789' + b'abcd'`` (copied into the temporary, which is then yielded)
    - ``b'abcdef'[4:6]``      --> output    ``b'ef'``            (slicing, no copy)
    """
    if output_fragment_size < 1:
        raise ValueError(f"Invalid output fragment size: {output_fragment_size}")

    carry: typing.Union[bytearray, memoryview] = memoryview(b"")
    for frag in input_fragments:
        # First, emit the leftover carry from the previous iteration(s), and update the fragment.
        # After this operation either the carry or the fragment (or both) will be empty.
        if carry:
            offset = output_fragment_size - len(carry)
            assert len(carry) < output_fragment_size and offset < output_fragment_size
            if isinstance(carry, bytearray):
                carry += frag[:offset]  # Expensive copy!
            else:
                carry = bytearray().join((carry, frag[:offset]))  # Expensive copy!

            frag = frag[offset:]
            if len(carry) >= output_fragment_size:
                assert len(carry) == output_fragment_size
                yield memoryview(carry)
                carry = memoryview(b"")

        assert not carry or not frag

        # Process the remaining data in the current fragment excepting the last incomplete section.
        for offset in range(0, len(frag), output_fragment_size):
            assert not carry
            chunk = frag[offset : offset + output_fragment_size]
            if len(chunk) < output_fragment_size:
                carry = chunk
            else:
                assert len(chunk) == output_fragment_size
                yield chunk

    if carry:
        assert len(carry) < output_fragment_size
        yield memoryview(carry)


def _unittest_util_refragment_manual() -> None:
    from pytest import raises

    with raises(ValueError):
        _ = list(refragment([memoryview(b"")], 0))

    assert [] == list(refragment([], 1000))
    assert [] == list(refragment([memoryview(b"")], 1000))

    def lby(it: typing.Iterable[memoryview]) -> typing.List[bytes]:
        return list(map(bytes, it))

    assert [b"012345"] == lby(refragment([memoryview(b"012345")], 1000))

    assert [b"0123456789"] == lby(refragment([memoryview(b"012345"), memoryview(b"6789")], 1000))
    assert [b"012345", b"6789"] == lby(refragment([memoryview(b"012345"), memoryview(b"6789")], 6))
    assert [b"012", b"345", b"678", b"9"] == lby(refragment([memoryview(b"012345"), memoryview(b"6789")], 3))
    assert [b"0", b"1", b"2", b"3", b"4", b"5", b"6", b"7", b"8", b"9"] == lby(
        refragment([memoryview(b"012345"), memoryview(b"6789"), memoryview(b"")], 1)
    )

    tiny = [
        memoryview(b"0"),
        memoryview(b"1"),
        memoryview(b"2"),
        memoryview(b"3"),
        memoryview(b"4"),
        memoryview(b"5"),
    ]
    assert [b"012345"] == lby(refragment(tiny, 1000))
    assert [b"0", b"1", b"2", b"3", b"4", b"5"] == lby(refragment(tiny, 1))


def _unittest_slow_util_refragment_automatic() -> None:
    import math
    import random

    def once(input_fragments: typing.List[memoryview], output_fragment_size: int) -> None:
        reference = _to_bytes(input_fragments)
        expected_frags = math.ceil(len(reference) / output_fragment_size)
        out = list(refragment(input_fragments, output_fragment_size))
        assert all(map(lambda x: isinstance(x, memoryview), out))
        assert len(out) == expected_frags
        assert _to_bytes(out) == reference
        if expected_frags > 0:
            sizes = list(map(len, out))
            assert all(x == output_fragment_size for x in sizes[:-1])
            assert 0 < sizes[-1] <= output_fragment_size

    def once_all(input_fragments: typing.List[memoryview]) -> None:
        longest = max(map(len, input_fragments)) if len(input_fragments) > 0 else 1
        for size in range(1, longest + 2):
            once(input_fragments, size)

        # Manual check for the edge case where all fragments are assembled into one chunk
        total_size = sum(map(len, input_fragments))
        if total_size > 0:
            out_list = list(refragment(input_fragments, total_size))
            assert len(out_list) in (0, 1)
            out = out_list[0] if out_list else b""
            assert out == _to_bytes(input_fragments)

    once_all([])
    once_all([memoryview(b"012345"), memoryview(b"6789")])

    num_iterations = 100
    max_fragments = 100
    max_fragment_size = 100

    def make_random_fragment() -> memoryview:
        size = random.randint(0, max_fragment_size)
        return memoryview(bytes(random.getrandbits(8) for _ in range(size)))

    for _ in range(num_iterations):
        num_fragments = random.randint(0, max_fragments)
        frags = [make_random_fragment() for _ in range(num_fragments)]
        once_all(frags)


def _to_bytes(fragments: typing.Iterable[memoryview]) -> bytes:
    return bytes().join(fragments)


def _unittest_util_refragment_to_bytes() -> None:
    assert _to_bytes([]) == b""
    assert _to_bytes([memoryview(b"")]) == b""
    assert _to_bytes([memoryview(b"")] * 3) == b""
    assert _to_bytes([memoryview(b""), memoryview(b"123"), memoryview(b"")]) == b"123"
    assert _to_bytes([memoryview(b"123")]) == b"123"
    assert _to_bytes([memoryview(b"123"), memoryview(b"456")]) == b"123456"
    assert _to_bytes([memoryview(b"123"), memoryview(b""), memoryview(b"456")]) == b"123456"



# Copyright (c) 2020 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import typing
from pyuavcan.transport import TransferFrom, Timestamp
from . import TransferReassembler, Frame


class AlienTransferReassembler:
    """
    This is a wrapper over :class:`TransferReassembler` optimized for tracing rather than real-time communication.
    It implements heuristics optimized for diagnostics and inspection rather than real-time operation.

    The caller is expected to keep a registry (dict) of session tracers indexed by their session specifiers,
    which are extracted from captured transport frames.
    """

    _MAX_INTERVAL = 1.0
    _TID_TIMEOUT_MULTIPLIER = 2.0  # TID = 2*interval as suggested in the Specification.

    _EXTENT_BYTES = 2 ** 32
    """
    The extent is effectively unlimited -- we want to be able to process all transfers.
    """

    def __init__(self, source_node_id: int) -> None:
        self._last_error: typing.Optional[TransferReassembler.Error] = None
        self._reassembler = TransferReassembler(
            source_node_id=source_node_id,
            extent_bytes=AlienTransferReassembler._EXTENT_BYTES,
            on_error_callback=self._register_reassembly_error,
        )
        self._last_transfer_monotonic: float = 0.0
        self._interval = float(AlienTransferReassembler._MAX_INTERVAL)

    def process_frame(
        self, timestamp: Timestamp, frame: Frame
    ) -> typing.Union[TransferFrom, TransferReassembler.Error, None]:
        trf = self._reassembler.process_frame(
            timestamp=timestamp, frame=frame, transfer_id_timeout=self.transfer_id_timeout
        )
        if trf is None:
            out, self._last_error = self._last_error, None
            return out

        # Update the transfer-ID timeout.
        delta = float(trf.timestamp.monotonic) - self._last_transfer_monotonic
        delta = min(AlienTransferReassembler._MAX_INTERVAL, max(0.0, delta))
        self._interval = (self._interval + delta) * 0.5
        self._last_transfer_monotonic = float(trf.timestamp.monotonic)

        return trf

    @property
    def transfer_id_timeout(self) -> float:
        """
        The current value of the auto-deduced transfer-ID timeout.
        It is automatically adjusted whenever a new transfer is received.
        """
        return self._interval * AlienTransferReassembler._TID_TIMEOUT_MULTIPLIER

    def _register_reassembly_error(self, error: TransferReassembler.Error) -> None:
        self._last_error = error



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

r"""
Submodule import policy
+++++++++++++++++++++++

The following submodules are auto-imported when the root module ``pyuavcan`` is imported:

- :mod:`pyuavcan.dsdl`
- :mod:`pyuavcan.transport`, but not concrete transport implementation submodules.
- :mod:`pyuavcan.presentation`
- :mod:`pyuavcan.util`

Submodule :mod:`pyuavcan.application` is not auto-imported because in order to have it imported
the DSDL-generated package ``uavcan`` containing the standard data types must be compiled first.


Log level override
++++++++++++++++++

The environment variable ``PYUAVCAN_LOGLEVEL`` can be set to one of the following values to override
the library log level:

- ``CRITICAL``
- ``FATAL``
- ``ERROR``
- ``WARNING``
- ``INFO``
- ``DEBUG``
"""

import os as _os
from importlib.resources import read_text as _read_text


__version__ = _read_text(__name__, "VERSION", encoding="utf8").strip()
__version_info__ = tuple(map(int, __version__.split(".")[:3]))
__author__ = "UAVCAN Consortium"
__copyright__ = "Copyright (c) 2019 UAVCAN Consortium"
__email__ = "consortium@uavcan.org"
__license__ = "MIT"


UAVCAN_SPECIFICATION_VERSION = 1, 0
"""
Version of the UAVCAN protocol implemented by this library, major and minor.
The corresponding field in ``uavcan.node.GetInfo.Response`` is initialized from this value,
see :func:`pyuavcan.application.make_node`.
"""


_log_level_from_env = _os.environ.get("PYUAVCAN_LOGLEVEL")
if _log_level_from_env is not None:
    import logging as _logging

    _logging.basicConfig(
        format="%(asctime)s %(process)5d %(levelname)-8s %(name)s: %(message)s", level=_log_level_from_env
    )
    _logging.getLogger(__name__).setLevel(_log_level_from_env)
    _logging.getLogger(__name__).info("Log config from env var; level: %r", _log_level_from_env)


# The sub-packages are imported in the order of their interdependency.
import pyuavcan.util as util  # pylint: disable=R0402,C0413  # noqa
import pyuavcan.dsdl as dsdl  # pylint: disable=R0402,C0413  # noqa
import pyuavcan.transport as transport  # pylint: disable=R0402,C0413  # noqa
import pyuavcan.presentation as presentation  # pylint: disable=R0402,C0413  # noqa



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import typing
import struct
import dataclasses
from cobs import cobs  # type: ignore
import pyuavcan

_VERSION = 0

# Same value represents broadcast node ID when transmitting.
_ANONYMOUS_NODE_ID = 0xFFFF

_HEADER_WITHOUT_CRC_FORMAT = struct.Struct(
    "<"
    "BB"  # Version, priority
    "HHH"  # source NID, destination NID, data specifier
    "8x"  # reserved 64 bits
    "Q"  # transfer-ID
    "L"
)  # Frame index with end-of-transfer flag in the MSB
_CRC_SIZE_BYTES = len(pyuavcan.transport.commons.high_overhead_transport.TransferCRC().value_as_bytes)
_HEADER_SIZE = _HEADER_WITHOUT_CRC_FORMAT.size + _CRC_SIZE_BYTES
assert _HEADER_SIZE == 32


@dataclasses.dataclass(frozen=True, repr=False)
class SerialFrame(pyuavcan.transport.commons.high_overhead_transport.Frame):
    NODE_ID_MASK = 4095
    TRANSFER_ID_MASK = 2 ** 64 - 1
    INDEX_MASK = 2 ** 31 - 1

    NODE_ID_RANGE = range(NODE_ID_MASK + 1)

    FRAME_DELIMITER_BYTE = 0x00

    NUM_OVERHEAD_BYTES_EXCEPT_DELIMITERS_AND_ESCAPING = _HEADER_SIZE + _CRC_SIZE_BYTES

    source_node_id: typing.Optional[int]
    destination_node_id: typing.Optional[int]
    data_specifier: pyuavcan.transport.DataSpecifier

    def __post_init__(self) -> None:
        if self.source_node_id is not None and not (0 <= self.source_node_id <= self.NODE_ID_MASK):
            raise ValueError(f"Invalid source node ID: {self.source_node_id}")

        if self.destination_node_id is not None and not (0 <= self.destination_node_id <= self.NODE_ID_MASK):
            raise ValueError(f"Invalid destination node ID: {self.destination_node_id}")

        if isinstance(self.data_specifier, pyuavcan.transport.ServiceDataSpecifier) and self.source_node_id is None:
            raise ValueError(f"Anonymous nodes cannot use service transfers: {self.data_specifier}")

        if not isinstance(self.data_specifier, pyuavcan.transport.DataSpecifier):
            raise TypeError(f"Invalid data specifier: {self.data_specifier}")  # pragma: no cover

        if not (0 <= self.transfer_id <= self.TRANSFER_ID_MASK):
            raise ValueError(f"Invalid transfer-ID: {self.transfer_id}")

        if not (0 <= self.index <= self.INDEX_MASK):
            raise ValueError(f"Invalid frame index: {self.index}")

    def compile_into(self, out_buffer: bytearray) -> memoryview:
        """
        Compiles the frame into the specified output buffer, escaping the data as necessary.
        The buffer must be large enough to accommodate the frame header with the payload and CRC,
        including escape sequences.
        :returns: View of the memory from the beginning of the buffer until the end of the compiled frame.
        """
        src_nid = _ANONYMOUS_NODE_ID if self.source_node_id is None else self.source_node_id
        dst_nid = _ANONYMOUS_NODE_ID if self.destination_node_id is None else self.destination_node_id

        if isinstance(self.data_specifier, pyuavcan.transport.MessageDataSpecifier):
            data_spec = self.data_specifier.subject_id
        elif isinstance(self.data_specifier, pyuavcan.transport.ServiceDataSpecifier):
            is_response = self.data_specifier.role == self.data_specifier.Role.RESPONSE
            data_spec = (1 << 15) | ((1 << 14) if is_response else 0) | self.data_specifier.service_id
        else:
            assert False

        index_eot = self.index | ((1 << 31) if self.end_of_transfer else 0)

        header = _HEADER_WITHOUT_CRC_FORMAT.pack(
            _VERSION, int(self.priority), src_nid, dst_nid, data_spec, self.transfer_id, index_eot
        )
        header += pyuavcan.transport.commons.crc.CRC32C.new(header).value_as_bytes
        assert len(header) == _HEADER_SIZE

        payload_crc_bytes = pyuavcan.transport.commons.crc.CRC32C.new(self.payload).value_as_bytes

        out_buffer[0] = self.FRAME_DELIMITER_BYTE
        next_byte_index = 1

        # noinspection PyTypeChecker
        packet_bytes = header + self.payload + payload_crc_bytes
        encoded_image = cobs.encode(packet_bytes)
        # place in the buffer and update next_byte_index:
        out_buffer[next_byte_index : next_byte_index + len(encoded_image)] = encoded_image
        next_byte_index += len(encoded_image)

        out_buffer[next_byte_index] = self.FRAME_DELIMITER_BYTE
        next_byte_index += 1

        assert (next_byte_index - 2) >= (len(header) + len(self.payload) + len(payload_crc_bytes))
        return memoryview(out_buffer)[:next_byte_index]

    @staticmethod
    def calc_cobs_size(payload_size_bytes: int) -> int:
        """
        :returns: worst case COBS-encoded message size for a given payload size.
        """
        # equivalent to int(math.ceil(payload_size_bytes * 255.0 / 254.0)
        return (payload_size_bytes * 255 + 253) // 254

    @staticmethod
    def parse_from_cobs_image(image: memoryview) -> typing.Optional[SerialFrame]:
        """
        Delimiters will be stripped if present but they are not required.
        :returns: Frame or None if the image is invalid.
        """
        try:
            while image[0] == SerialFrame.FRAME_DELIMITER_BYTE:
                image = image[1:]
            while image[-1] == SerialFrame.FRAME_DELIMITER_BYTE:
                image = image[:-1]
        except IndexError:
            return None
        try:
            unescaped_image = cobs.decode(bytearray(image))  # TODO: PERFORMANCE WARNING: AVOID THE COPY
        except cobs.DecodeError:
            return None
        return SerialFrame.parse_from_unescaped_image(memoryview(unescaped_image))

    @staticmethod
    def parse_from_unescaped_image(header_payload_crc_image: memoryview) -> typing.Optional[SerialFrame]:
        """
        :returns: Frame or None if the image is invalid.
        """
        if len(header_payload_crc_image) < SerialFrame.NUM_OVERHEAD_BYTES_EXCEPT_DELIMITERS_AND_ESCAPING:
            return None

        header = header_payload_crc_image[:_HEADER_SIZE]
        if not pyuavcan.transport.commons.crc.CRC32C.new(header).check_residue():
            return None

        payload_with_crc = header_payload_crc_image[_HEADER_SIZE:]
        if not pyuavcan.transport.commons.crc.CRC32C.new(payload_with_crc).check_residue():
            return None
        payload = payload_with_crc[:-_CRC_SIZE_BYTES]

        # noinspection PyTypeChecker
        (
            version,
            int_priority,
            src_nid,
            dst_nid,
            int_data_spec,
            transfer_id,
            index_eot,
        ) = _HEADER_WITHOUT_CRC_FORMAT.unpack_from(header)
        if version != _VERSION:
            return None

        src_nid = None if src_nid == _ANONYMOUS_NODE_ID else src_nid
        dst_nid = None if dst_nid == _ANONYMOUS_NODE_ID else dst_nid

        try:  # https://github.com/UAVCAN/pyuavcan/issues/176
            data_specifier: pyuavcan.transport.DataSpecifier
            if int_data_spec & (1 << 15) == 0:
                data_specifier = pyuavcan.transport.MessageDataSpecifier(int_data_spec)
            else:
                if int_data_spec & (1 << 14):
                    role = pyuavcan.transport.ServiceDataSpecifier.Role.RESPONSE
                else:
                    role = pyuavcan.transport.ServiceDataSpecifier.Role.REQUEST
                service_id = int_data_spec & pyuavcan.transport.ServiceDataSpecifier.SERVICE_ID_MASK
                data_specifier = pyuavcan.transport.ServiceDataSpecifier(service_id, role)

            # noinspection PyArgumentList
            return SerialFrame(
                priority=pyuavcan.transport.Priority(int_priority),
                source_node_id=src_nid,
                destination_node_id=dst_nid,
                data_specifier=data_specifier,
                transfer_id=transfer_id,
                index=index_eot & SerialFrame.INDEX_MASK,
                end_of_transfer=index_eot & (1 << 31) != 0,
                payload=payload,
            )
        except ValueError:
            return None


# ----------------------------------------  TESTS GO BELOW THIS LINE  ----------------------------------------


def _unittest_frame_compile_message() -> None:
    from pyuavcan.transport import Priority, MessageDataSpecifier

    f = SerialFrame(
        priority=Priority.HIGH,
        source_node_id=SerialFrame.FRAME_DELIMITER_BYTE,
        destination_node_id=SerialFrame.FRAME_DELIMITER_BYTE,
        data_specifier=MessageDataSpecifier(2345),
        transfer_id=1234567890123456789,
        index=1234567,
        end_of_transfer=True,
        payload=memoryview(b"abcd\x00ef\x00"),
    )

    buffer = bytearray(0 for _ in range(1000))
    mv = f.compile_into(buffer)

    assert mv[0] == SerialFrame.FRAME_DELIMITER_BYTE
    assert mv[-1] == SerialFrame.FRAME_DELIMITER_BYTE

    segment_cobs = bytes(mv[1:-1])
    assert SerialFrame.FRAME_DELIMITER_BYTE not in segment_cobs

    segment = cobs.decode(segment_cobs)

    # Header validation
    assert segment[0] == _VERSION
    assert segment[1] == int(Priority.HIGH)
    assert (segment[2], segment[3]) == (SerialFrame.FRAME_DELIMITER_BYTE, 0)
    assert (segment[4], segment[5]) == (SerialFrame.FRAME_DELIMITER_BYTE, 0)
    assert segment[6:8] == 2345 .to_bytes(2, "little")
    assert segment[8:16] == b"\x00" * 8
    assert segment[16:24] == 1234567890123456789 .to_bytes(8, "little")
    assert segment[24:28] == (1234567 + 0x8000_0000).to_bytes(4, "little")
    # Header CRC here

    # Payload validation
    assert segment[32:40] == b"abcd\x00ef\x00"
    assert segment[40:] == pyuavcan.transport.commons.crc.CRC32C.new(f.payload).value_as_bytes


def _unittest_frame_compile_service() -> None:
    from pyuavcan.transport import Priority, ServiceDataSpecifier

    f = SerialFrame(
        priority=Priority.FAST,
        source_node_id=SerialFrame.FRAME_DELIMITER_BYTE,
        destination_node_id=None,
        data_specifier=ServiceDataSpecifier(123, ServiceDataSpecifier.Role.RESPONSE),
        transfer_id=1234567890123456789,
        index=1234567,
        end_of_transfer=False,
        payload=memoryview(b""),
    )

    buffer = bytearray(0 for _ in range(50))
    mv = f.compile_into(buffer)

    assert mv[0] == mv[-1] == SerialFrame.FRAME_DELIMITER_BYTE
    segment_cobs = bytes(mv[1:-1])
    assert SerialFrame.FRAME_DELIMITER_BYTE not in segment_cobs

    segment = cobs.decode(segment_cobs)

    # Header validation
    assert segment[0] == _VERSION
    assert segment[1] == int(Priority.FAST)
    assert (segment[2], segment[3]) == (SerialFrame.FRAME_DELIMITER_BYTE, 0)
    assert (segment[4], segment[5]) == (0xFF, 0xFF)
    assert segment[6:8] == ((1 << 15) | (1 << 14) | 123).to_bytes(2, "little")
    assert segment[8:16] == b"\x00" * 8
    assert segment[16:24] == 1234567890123456789 .to_bytes(8, "little")
    assert segment[24:28] == 1234567 .to_bytes(4, "little")
    # Header CRC here

    # CRC validation
    assert segment[32:] == pyuavcan.transport.commons.crc.CRC32C.new(f.payload).value_as_bytes


def _unittest_frame_parse() -> None:
    from pyuavcan.transport import Priority, MessageDataSpecifier, ServiceDataSpecifier

    def get_crc(*blocks: typing.Union[bytes, memoryview]) -> bytes:
        return pyuavcan.transport.commons.crc.CRC32C.new(*blocks).value_as_bytes

    # Valid message with payload
    header = bytes(
        [
            _VERSION,
            int(Priority.LOW),
            0x7B,
            0x00,  # Source NID        123
            0xC8,
            0x01,  # Destination NID   456
            0xE1,
            0x10,  # Data specifier    4321
            0x00,
            0x00,
            0x00,
            0x00,
            0x00,
            0x00,
            0x00,
            0x00,  # Reserved
            0xD2,
            0x0A,
            0x1F,
            0xEB,
            0x8C,
            0xA9,
            0x54,
            0xAB,  # Transfer ID       12345678901234567890
            0x31,
            0xD4,
            0x00,
            0x80,  # Frame index, EOT  54321 with EOT flag set
        ]
    )
    header += get_crc(header)
    assert len(header) == 32
    payload = b"Squeeze mayonnaise onto a hamster"
    f = SerialFrame.parse_from_unescaped_image(memoryview(header + payload + get_crc(payload)))
    assert f == SerialFrame(
        priority=Priority.LOW,
        source_node_id=123,
        destination_node_id=456,
        data_specifier=MessageDataSpecifier(4321),
        transfer_id=12345678901234567890,
        index=54321,
        end_of_transfer=True,
        payload=memoryview(payload),
    )

    # Valid service with no payload
    header = bytes(
        [
            _VERSION,
            int(Priority.LOW),
            0x01,
            0x00,
            0x00,
            0x00,
            0x10,
            0xC0,  # Response, service ID 16
            0x00,
            0x00,
            0x00,
            0x00,
            0x00,
            0x00,
            0x00,
            0x00,
            0xD2,
            0x0A,
            0x1F,
            0xEB,
            0x8C,
            0xA9,
            0x54,
            0xAB,
            0x31,
            0xD4,
            0x00,
            0x00,
        ]
    )
    header += get_crc(header)
    assert len(header) == 32
    f = SerialFrame.parse_from_unescaped_image(memoryview(header + get_crc(b"")))
    assert f == SerialFrame(
        priority=Priority.LOW,
        source_node_id=1,
        destination_node_id=0,
        data_specifier=ServiceDataSpecifier(16, ServiceDataSpecifier.Role.RESPONSE),
        transfer_id=12345678901234567890,
        index=54321,
        end_of_transfer=False,
        payload=memoryview(b""),
    )

    # Valid service with no payload
    header = bytes(
        [
            _VERSION,
            int(Priority.LOW),
            0x01,
            0x00,
            0x00,
            0x00,
            0x10,
            0x80,  # Request, service ID 16
            0x00,
            0x00,
            0x00,
            0x00,
            0x00,
            0x00,
            0x00,
            0x00,
            0xD2,
            0x0A,
            0x1F,
            0xEB,
            0x8C,
            0xA9,
            0x54,
            0xAB,
            0x31,
            0xD4,
            0x00,
            0x00,
        ]
    )
    header += get_crc(header)
    assert len(header) == 32
    f = SerialFrame.parse_from_unescaped_image(memoryview(header + get_crc(b"")))
    assert f == SerialFrame(
        priority=Priority.LOW,
        source_node_id=1,
        destination_node_id=0,
        data_specifier=ServiceDataSpecifier(16, ServiceDataSpecifier.Role.REQUEST),
        transfer_id=12345678901234567890,
        index=54321,
        end_of_transfer=False,
        payload=memoryview(b""),
    )

    # Too short
    assert SerialFrame.parse_from_unescaped_image(memoryview(header[1:] + get_crc(payload))) is None

    # Bad CRC
    assert SerialFrame.parse_from_unescaped_image(memoryview(header + payload + b"1234")) is None

    # Bad version
    header = bytes(
        [
            _VERSION + 1,
            int(Priority.LOW),
            0xFF,
            0xFF,
            0x00,
            0x00,
            0xE1,
            0x10,
            0x00,
            0x00,
            0x00,
            0x00,
            0x00,
            0x00,
            0x00,
            0x00,
            0xD2,
            0x0A,
            0x1F,
            0xEB,
            0x8C,
            0xA9,
            0x54,
            0xAB,
            0x31,
            0xD4,
            0x00,
            0x00,
        ]
    )
    header += get_crc(header)
    assert len(header) == 32
    assert SerialFrame.parse_from_unescaped_image(memoryview(header + get_crc(b""))) is None

    # Bad fields
    header = bytes(
        [
            _VERSION,
            0x88,
            0xFF,
            0xFF,
            0x00,
            0xFF,
            0xE1,
            0x10,
            0x00,
            0x00,
            0x00,
            0x00,
            0x00,
            0x00,
            0x00,
            0x00,
            0xD2,
            0x0A,
            0x1F,
            0xEB,
            0x8C,
            0xA9,
            0x54,
            0xAB,
            0x31,
            0xD4,
            0x00,
            0x00,
        ]
    )
    header += get_crc(header)
    assert len(header) == 32
    assert SerialFrame.parse_from_unescaped_image(memoryview(header + get_crc(b""))) is None


def _unittest_frame_check() -> None:
    from pytest import raises
    from pyuavcan.transport import Priority, MessageDataSpecifier, ServiceDataSpecifier

    _ = SerialFrame(
        priority=Priority.HIGH,
        source_node_id=123,
        destination_node_id=456,
        data_specifier=MessageDataSpecifier(2345),
        transfer_id=1234567890123456789,
        index=1234567,
        end_of_transfer=False,
        payload=memoryview(b"abcdef"),
    )

    with raises(ValueError):
        SerialFrame(
            priority=Priority.HIGH,
            source_node_id=123456,
            destination_node_id=456,
            data_specifier=MessageDataSpecifier(2345),
            transfer_id=1234567890123456789,
            index=1234567,
            end_of_transfer=False,
            payload=memoryview(b"abcdef"),
        )

    with raises(ValueError):
        SerialFrame(
            priority=Priority.HIGH,
            source_node_id=123,
            destination_node_id=123456,
            data_specifier=MessageDataSpecifier(2345),
            transfer_id=1234567890123456789,
            index=1234567,
            end_of_transfer=False,
            payload=memoryview(b"abcdef"),
        )

    with raises(ValueError):
        SerialFrame(
            priority=Priority.HIGH,
            source_node_id=None,
            destination_node_id=456,
            data_specifier=ServiceDataSpecifier(123, ServiceDataSpecifier.Role.REQUEST),
            transfer_id=1234567890123456789,
            index=1234567,
            end_of_transfer=False,
            payload=memoryview(b"abcdef"),
        )

    with raises(ValueError):
        SerialFrame(
            priority=Priority.HIGH,
            source_node_id=None,
            destination_node_id=None,
            data_specifier=MessageDataSpecifier(2345),
            transfer_id=-1,
            index=1234567,
            end_of_transfer=False,
            payload=memoryview(b"abcdef"),
        )

    with raises(ValueError):
        SerialFrame(
            priority=Priority.HIGH,
            source_node_id=None,
            destination_node_id=None,
            data_specifier=MessageDataSpecifier(2345),
            transfer_id=0,
            index=-1,
            end_of_transfer=False,
            payload=memoryview(b"abcdef"),
        )



# Copyright (C) 2021  UAVCAN Consortium  <uavcan.org>
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
from typing import Union, Optional, Iterator, Any
from pathlib import Path
import logging
import sqlite3
import pyuavcan
from . import Entry, BackendError, Backend, Value


__all__ = ["StaticBackend"]


_TIMEOUT = 0.5
_LOCATION_VOLATILE = ":memory:"


# noinspection SqlNoDataSourceInspection,SqlResolve
class StaticBackend(Backend):
    """
    Register storage backend implementation based on SQLite.
    Supports either persistent on-disk single-file storage or volatile in-memory storage.

    >>> b = StaticBackend("my_register_file.db")
    >>> b.persistent    # If a file is specified, the storage is persistent.
    True
    >>> b.location
    'my_register_file.db'
    >>> b.close()
    >>> b = StaticBackend()
    >>> b.persistent    # If no file is specified, the data is kept in-memory.
    False
    >>> from pyuavcan.application.register import Bit
    >>> b["foo"] = Value(bit=Bit([True, False, True]))   # Create new register.
    >>> b["foo"].mutable
    True
    >>> list(b["foo"].value.bit.value)
    [True, False, True]
    >>> b["foo"] = Value(bit=Bit([False, True, True]))  # Set new value.
    >>> list(b["foo"].value.bit.value)
    [False, True, True]
    >>> list(b)
    ['foo']
    >>> del b["foo"]
    >>> list(b)
    []
    """

    def __init__(self, location: Union[None, str, Path] = None):
        """
        :param location: Either a path to the database file, or None. If None, the data will be stored in memory.

        The database is always initialized with ``check_same_thread=False`` to enable delegating its initialization
        to a thread pool from an async context.
        This is important for this library because if one needs to initialize a new node from an async function,
        calling the factories directly may be unacceptable due to their blocking behavior,
        so one is likely to rely on :meth:`asyncio.loop.run_in_executor`.
        The executor will initialize the instance in a worker thread and then hand it over to the main thread,
        which is perfectly safe, but it would trigger a false error from the SQLite engine complaining about
        the possibility of concurrency-related bugs.
        """
        self._loc = str(location or _LOCATION_VOLATILE).strip()
        self._db = sqlite3.connect(self._loc, timeout=_TIMEOUT, check_same_thread=False)
        self._execute(
            r"""
            create table if not exists `register` (
                `name`      varchar(255) not null unique primary key,
                `value`     blob not null,
                `mutable`   boolean not null,
                `ts`        time not null default current_timestamp
            )
            """,
            commit=True,
        )
        _logger.debug("%r: Initialized with registers: %r", self, self.keys())
        super().__init__()

    @property
    def location(self) -> str:
        return self._loc

    @property
    def persistent(self) -> bool:
        return self._loc.lower() != _LOCATION_VOLATILE

    def close(self) -> None:
        self._db.close()

    def index(self, index: int) -> Optional[str]:
        res = self._execute(r"select name from register order by name limit 1 offset ?", index).fetchone()
        return res[0] if res else None

    def setdefault(self, key: str, default: Optional[Union[Entry, Value]] = None) -> Entry:
        # This override is necessary to support assignment of Value along with Entry.
        if key not in self:
            if default is None:
                raise TypeError  # pragma: no cover
            self[key] = default
        return self[key]

    def __getitem__(self, key: str) -> Entry:
        res = self._execute(r"select mutable, value from register where name = ?", key).fetchone()
        if res is None:
            raise KeyError(key)
        mutable, value = res
        assert isinstance(value, bytes)
        obj = pyuavcan.dsdl.deserialize(Value, [memoryview(value)])
        if obj is None:  # pragma: no cover
            _logger.warning("%r: Value of %r is not a valid serialization of %s: %r", self, key, Value, value)
            raise KeyError(key)
        e = Entry(value=obj, mutable=bool(mutable))
        _logger.debug("%r: Get %r -> %r", self, key, e)
        return e

    def __setitem__(self, key: str, value: Union[Entry, Value]) -> None:
        """
        If the register does not exist, it will be implicitly created.
        If the value is an instance of :class:`Value`, the mutability flag defaults to the old value or True if none.
        """
        if isinstance(value, Value):
            try:
                mutable = self[key].mutable
            except KeyError:
                mutable = True
            e = Entry(value, mutable=mutable)
        elif isinstance(value, Entry):
            e = value
        else:  # pragma: no cover
            raise TypeError(f"Unexpected argument: {value!r}")
        _logger.debug("%r: Set %r <- %r", self, key, e)
        # language=SQLite
        self._execute(
            r"insert or replace into register (name, value, mutable) values (?, ?, ?)",
            key,
            b"".join(pyuavcan.dsdl.serialize(e.value)),
            e.mutable,
            commit=True,
        )

    def __delitem__(self, key: str) -> None:
        _logger.debug("%r: Delete %r", self, key)
        self._execute(r"delete from register where name = ?", key, commit=True)

    def __iter__(self) -> Iterator[str]:
        return iter(x for x, in self._execute(r"select name from register order by name").fetchall())

    def __len__(self) -> int:
        return int(self._execute(r"select count(*) from register").fetchone()[0])

    def _execute(self, statement: str, *params: Any, commit: bool = False) -> sqlite3.Cursor:
        try:
            cur = self._db.execute(statement, params)
            if commit:
                self._db.commit()
            return cur
        except sqlite3.OperationalError as ex:
            raise BackendError(f"Database transaction has failed: {ex}") from ex


_logger = logging.getLogger(__name__)


def _unittest_memory() -> None:
    from uavcan.primitive import String_1 as String, Unstructured_1 as Unstructured

    st = StaticBackend()
    print(st)
    assert not st.keys()
    assert not st.index(0)
    assert None is st.get("foo")
    assert len(st) == 0
    del st["foo"]

    st["foo"] = Value(string=String("Hello world!"))
    e = st.get("foo")
    assert e
    assert e.value.string
    assert e.value.string.value.tobytes().decode() == "Hello world!"
    assert e.mutable
    assert len(st) == 1

    # Override the same register.
    st["foo"] = Value(unstructured=Unstructured([1, 2, 3]))
    e = st.get("foo")
    assert e
    assert e.value.unstructured
    assert e.value.unstructured.value.tobytes() == b"\x01\x02\x03"
    assert e.mutable
    assert len(st) == 1

    assert ["foo"] == list(st.keys())
    assert "foo" == st.index(0)
    assert None is st.index(1)
    assert ["foo"] == list(st.keys())
    del st["foo"]
    assert [] == list(st.keys())
    assert len(st) == 0

    st.close()


def _unittest_file() -> None:
    import tempfile
    from uavcan.primitive import Unstructured_1 as Unstructured

    # First, populate the database with registers.
    db_file = tempfile.mktemp(".db")
    print("DB file:", db_file)
    st = StaticBackend(db_file)
    print(st)
    st["a"] = Value(unstructured=Unstructured([1, 2, 3]))
    st["b"] = Value(unstructured=Unstructured([4, 5, 6]))
    assert len(st) == 2
    st.close()

    # Then re-open it in writeable mode and ensure correctness.
    st = StaticBackend(db_file)
    print(st)
    assert len(st) == 2
    e = st.get("a")
    assert e
    assert e.value.unstructured
    assert e.value.unstructured.value.tobytes() == b"\x01\x02\x03"
    assert e.mutable

    e = st.get("b")
    assert e
    assert e.value.unstructured
    assert e.value.unstructured.value.tobytes() == b"\x04\x05\x06"
    assert e.mutable
    st.close()



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import typing
import logging
import numpy
from numpy.typing import NDArray
import pydsdl
from ._composite_object import CompositeObject, get_model, get_attribute, set_attribute, get_class
from ._composite_object import CompositeObjectTypeVar


def to_builtin(obj: CompositeObject) -> typing.Dict[str, typing.Any]:
    """
    Accepts a DSDL object (an instance of a Python class auto-generated from a DSDL definition),
    returns its value represented using only native built-in types: dict, list, bool, int, float, str.
    Ordering of dict elements is guaranteed to match the field ordering of the source definition.
    Keys of dicts representing DSDL objects use the original unstropped names from the source DSDL definition;
    e.g., ``if``, not ``if_``.

    This is intended for use with JSON, YAML, and other serialization formats.

    ..  doctest::
        :hide:

        >>> import tests
        >>> _ = tests.dsdl.compile()

    >>> import json
    >>> import uavcan.primitive.array
    >>> json.dumps(to_builtin(uavcan.primitive.array.Integer32_1_0([-123, 456, 0])))
    '{"value": [-123, 456, 0]}'
    >>> import uavcan.register
    >>> request = uavcan.register.Access_1_0.Request(
    ...     uavcan.register.Name_1_0('my.register'),
    ...     uavcan.register.Value_1_0(integer16=uavcan.primitive.array.Integer16_1_0([1, 2, +42, -10_000]))
    ... )
    >>> to_builtin(request)  # doctest: +NORMALIZE_WHITESPACE
    {'name':  {'name': 'my.register'},
     'value': {'integer16': {'value': [1, 2, 42, -10000]}}}
    """
    model = get_model(obj)
    _raise_if_service_type(model)
    out = _to_builtin_impl(obj, model)
    assert isinstance(out, dict)
    return out


def _to_builtin_impl(
    obj: typing.Union[CompositeObject, NDArray[typing.Any], str, bool, int, float], model: pydsdl.SerializableType
) -> typing.Union[typing.Dict[str, typing.Any], typing.List[typing.Any], str, bool, int, float]:
    if isinstance(model, pydsdl.CompositeType):
        assert isinstance(obj, CompositeObject)
        return {
            f.name: _to_builtin_impl(get_attribute(obj, f.name), f.data_type)
            for f in model.fields_except_padding
            if get_attribute(obj, f.name) is not None  # The check is to hide inactive union variants.
        }

    if isinstance(model, pydsdl.ArrayType):
        assert isinstance(obj, numpy.ndarray)
        if model.string_like:  # TODO: drop this special case when strings are natively supported in DSDL.
            try:
                return bytes(e for e in obj).decode()
            except UnicodeError:
                return list(map(int, obj))
        return [_to_builtin_impl(e, model.element_type) for e in obj]

    if isinstance(model, pydsdl.PrimitiveType):
        # The explicit conversions are needed to get rid of NumPy scalar types.
        if isinstance(model, pydsdl.IntegerType):
            return int(obj)  # type: ignore
        if isinstance(model, pydsdl.FloatType):
            return float(obj)  # type: ignore
        if isinstance(model, pydsdl.BooleanType):
            return bool(obj)
        assert isinstance(obj, str)
        return obj

    assert False, "Unexpected inputs"


def update_from_builtin(destination: CompositeObjectTypeVar, source: typing.Any) -> CompositeObjectTypeVar:
    """
    Updates the provided DSDL object (an instance of a Python class auto-generated from a DSDL definition)
    with the values from a native representation, where DSDL objects are represented as dicts, arrays
    are lists, and primitives are represented as int/float/bool. This is the reverse of :func:`to_builtin`.
    Values that are not specified in the source are not updated (left at their original values),
    so an empty source will leave the input object unchanged.

    Source field names shall match the original unstropped names provided in the DSDL definition;
    e.g., `if`, not `if_`. If there is more than one variant specified for a union type, the last
    specified variant takes precedence.
    If the structure of the source does not match the destination object, the correct representation
    may be deduced automatically as long as it can be done unambiguously.

    :param destination: The object to update. The update will be done in-place. If you don't want the source
        object modified, clone it beforehand.

    :param source: The :class:`dict` instance containing the values to update the destination object with.

    :return: A reference to destination (not a copy).

    :raises: :class:`ValueError` if the provided source values cannot be applied to the destination object,
        or if the source contains fields that are not present in the destination object.
        :class:`TypeError` if an entity of the source cannot be converted into the type expected by the destination.

    >>> import tests; tests.dsdl.compile()  # DSDL package generation not shown in this example.
    [...]
    >>> import json
    >>> import uavcan.primitive.array
    >>> import uavcan.register
    >>> request = uavcan.register.Access_1_0.Request(
    ...     uavcan.register.Name_1_0('my.register'),
    ...     uavcan.register.Value_1_0(string=uavcan.primitive.String_1_0('Hello world!'))
    ... )
    >>> request
    uavcan.register.Access.Request...name='my.register'...value='Hello world!'...
    >>> update_from_builtin(request, {  # Switch the Value union from string to int16; keep the name unchanged.
    ...     'value': {
    ...         'integer16': {
    ...             'value': [1, 2, 42, -10000]
    ...         }
    ...     }
    ... })  # doctest: +NORMALIZE_WHITESPACE
    uavcan.register.Access.Request...name='my.register'...value=[ 1, 2, 42,-10000]...

    The following examples showcase positional initialization:

    >>> from uavcan.node import Heartbeat_1
    >>> update_from_builtin(Heartbeat_1(), [123456, 1, 2])  # doctest: +NORMALIZE_WHITESPACE
    uavcan.node.Heartbeat.1.0(uptime=123456,
                              health=uavcan.node.Health.1.0(value=1),
                              mode=uavcan.node.Mode.1.0(value=2),
                              vendor_specific_status_code=0)
    >>> update_from_builtin(Heartbeat_1(), 123456)  # doctest: +NORMALIZE_WHITESPACE
    uavcan.node.Heartbeat.1.0(uptime=123456,
                              health=uavcan.node.Health.1.0(value=0),
                              mode=uavcan.node.Mode.1.0(value=0),
                              vendor_specific_status_code=0)
    >>> update_from_builtin(Heartbeat_1(), [0, 0, 0, 0, 0])  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
        ...
    TypeError: ...

    >>> update_from_builtin(uavcan.primitive.array.Real64_1(), 123.456) # doctest: +NORMALIZE_WHITESPACE
    uavcan.primitive.array.Real64.1.0(value=[123.456])
    >>> update_from_builtin(uavcan.primitive.array.Real64_1(), [123.456]) # doctest: +NORMALIZE_WHITESPACE
    uavcan.primitive.array.Real64.1.0(value=[123.456])
    >>> update_from_builtin(uavcan.primitive.array.Real64_1(), [123.456, -9]) # doctest: +NORMALIZE_WHITESPACE
    uavcan.primitive.array.Real64.1.0(value=[123.456, -9. ])

    >>> update_from_builtin(uavcan.register.Access_1_0.Request(), ["X", {"integer8": 99}])  # Same as the next one!
    uavcan.register.Access.Request...name='X'...value=[99]...
    >>> update_from_builtin(uavcan.register.Access_1_0.Request(), {'name': 'X', 'value': {'integer8': {'value': [99]}}})
    uavcan.register.Access.Request...name='X'...value=[99]...
    """
    _logger.debug("update_from_builtin: destination/source on the next lines:\n%r\n%r", destination, source)
    if not isinstance(destination, CompositeObject):  # pragma: no cover
        raise TypeError(f"Bad destination: expected a CompositeObject, got {type(destination).__name__}")
    model = get_model(destination)
    _raise_if_service_type(model)
    fields = model.fields_except_padding

    # UX improvement: https://github.com/UAVCAN/pyuavcan/issues/147 -- match the source against the data type.
    if not isinstance(source, dict):
        if not isinstance(source, (list, tuple)):  # Assume positional initialization.
            source = (source,)
        can_propagate = fields and isinstance(fields[0].data_type, (pydsdl.ArrayType, pydsdl.CompositeType))
        too_many_values = len(source) > (1 if isinstance(model.inner_type, pydsdl.UnionType) else len(fields))
        if can_propagate and too_many_values:
            _logger.debug(
                "update_from_builtin: %d source values cannot be applied to %s but the first field accepts "
                "positional initialization -- propagating down",
                len(source),
                type(destination).__name__,
            )
            source = [source]
        if len(source) > len(fields):
            raise TypeError(
                f"Cannot apply {len(source)} values to {len(fields)} fields in {type(destination).__name__}"
            )
        source = {f.name: v for f, v in zip(fields, source)}
        return update_from_builtin(destination, source)

    source = dict(source)  # Create copy to prevent mutation of the original

    for f in fields:
        field_type = f.data_type
        try:
            value = source.pop(f.name)
        except LookupError:
            continue  # No value specified, keep original value

        if isinstance(field_type, pydsdl.CompositeType):
            field_obj = get_attribute(destination, f.name)
            if field_obj is None:  # Oh, this is a union
                field_obj = get_class(field_type)()  # The variant was not selected, construct a default
                set_attribute(destination, f.name, field_obj)  # Switch the union to the new variant
            update_from_builtin(field_obj, value)

        elif isinstance(field_type, pydsdl.ArrayType):
            element_type = field_type.element_type
            if isinstance(element_type, pydsdl.PrimitiveType):
                set_attribute(destination, f.name, value)
            elif isinstance(element_type, pydsdl.CompositeType):
                dtype = get_class(element_type)
                set_attribute(destination, f.name, [update_from_builtin(dtype(), s) for s in value])
            else:
                assert False, f"Unexpected array element type: {element_type!r}"

        elif isinstance(field_type, pydsdl.PrimitiveType):
            set_attribute(destination, f.name, value)

        else:
            assert False, f"Unexpected field type: {field_type!r}"

    if source:
        raise ValueError(f"No such fields in {model}: {list(source.keys())}")

    return destination


def _raise_if_service_type(model: pydsdl.SerializableType) -> None:
    if isinstance(model, pydsdl.ServiceType):  # pragma: no cover
        raise TypeError(
            f"Built-in form is not defined for service types. "
            f"Did you mean to use Request or Response? Input type: {model}"
        )


_logger = logging.getLogger(__name__)



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import typing
from pyuavcan.transport import MessageDataSpecifier, ServiceDataSpecifier, InputSessionSpecifier
from ._session import CANInputSession
from ._identifier import CANID


class InputDispatchTable:
    """
    Time-memory trade-off: the input dispatch table is tens of megabytes large, but the lookup is very fast and O(1).
    This is necessary to ensure scalability for high-load applications such as real-time network monitoring.
    """

    _NUM_SUBJECTS = MessageDataSpecifier.SUBJECT_ID_MASK + 1
    _NUM_SERVICES = ServiceDataSpecifier.SERVICE_ID_MASK + 1
    _NUM_NODE_IDS = CANID.NODE_ID_MASK + 1

    # Services multiplied by two to account for requests and responses.
    # One added to nodes to allow promiscuous inputs which don't care about source node ID.
    _TABLE_SIZE = (_NUM_SUBJECTS + _NUM_SERVICES * 2) * (_NUM_NODE_IDS + 1)

    def __init__(self) -> None:
        # This method of construction is an order of magnitude faster than range-based. It matters here. A lot.
        self._table: typing.List[typing.Optional[CANInputSession]] = [None] * (self._TABLE_SIZE + 1)

        # A parallel dict is necessary for constant-complexity element listing. Traversing the table takes forever.
        self._dict: typing.Dict[InputSessionSpecifier, CANInputSession] = {}

    @property
    def items(self) -> typing.Iterable[CANInputSession]:
        return self._dict.values()

    def add(self, session: CANInputSession) -> None:
        """
        This method is used only when a new input session is created; performance is not a priority.
        """
        key = session.specifier
        self._table[self._compute_index(key)] = session
        self._dict[key] = session

    def get(self, specifier: InputSessionSpecifier) -> typing.Optional[CANInputSession]:
        """
        Constant-time lookup. Invoked for every received frame.
        """
        return self._table[self._compute_index(specifier)]

    def remove(self, specifier: InputSessionSpecifier) -> None:
        """
        This method is used only when an input session is destroyed; performance is not a priority.
        """
        self._table[self._compute_index(specifier)] = None
        del self._dict[specifier]

    @staticmethod
    def _compute_index(specifier: InputSessionSpecifier) -> int:
        ds, nid = specifier.data_specifier, specifier.remote_node_id
        if isinstance(ds, MessageDataSpecifier):
            dim1 = ds.subject_id
        elif isinstance(ds, ServiceDataSpecifier):
            if ds.role == ds.Role.REQUEST:
                dim1 = ds.service_id + InputDispatchTable._NUM_SUBJECTS
            elif ds.role == ds.Role.RESPONSE:
                dim1 = ds.service_id + InputDispatchTable._NUM_SUBJECTS + InputDispatchTable._NUM_SERVICES
            else:
                assert False
        else:
            assert False

        dim2_cardinality = InputDispatchTable._NUM_NODE_IDS + 1
        dim2 = nid if nid is not None else InputDispatchTable._NUM_NODE_IDS

        point = dim1 * dim2_cardinality + dim2

        assert 0 <= point < InputDispatchTable._TABLE_SIZE
        return point


def _unittest_input_dispatch_table() -> None:
    from pytest import raises
    from pyuavcan.transport import PayloadMetadata

    t = InputDispatchTable()
    assert len(list(t.items)) == 0
    assert t.get(InputSessionSpecifier(MessageDataSpecifier(1234), None)) is None
    with raises(LookupError):
        t.remove(InputSessionSpecifier(MessageDataSpecifier(1234), 123))

    a = CANInputSession(
        InputSessionSpecifier(MessageDataSpecifier(1234), None),
        PayloadMetadata(456),
        lambda: None,
    )
    t.add(a)
    t.add(a)
    assert list(t.items) == [a]
    assert t.get(InputSessionSpecifier(MessageDataSpecifier(1234), None)) == a
    t.remove(InputSessionSpecifier(MessageDataSpecifier(1234), None))
    assert len(list(t.items)) == 0


def _unittest_slow_input_dispatch_table_index() -> None:
    table_size = InputDispatchTable._TABLE_SIZE  # pylint: disable=protected-access
    values: typing.Set[int] = set()
    for node_id in (*range(InputDispatchTable._NUM_NODE_IDS), None):  # pylint: disable=protected-access
        for subj in range(InputDispatchTable._NUM_SUBJECTS):  # pylint: disable=protected-access
            out = InputDispatchTable._compute_index(  # pylint: disable=protected-access
                InputSessionSpecifier(MessageDataSpecifier(subj), node_id)
            )
            assert out not in values
            values.add(out)
            assert out < table_size

        for serv in range(InputDispatchTable._NUM_SERVICES):  # pylint: disable=protected-access
            for role in ServiceDataSpecifier.Role:
                out = InputDispatchTable._compute_index(  # pylint: disable=protected-access
                    InputSessionSpecifier(ServiceDataSpecifier(serv, role), node_id)
                )
                assert out not in values
                values.add(out)
                assert out < table_size

    assert len(values) == table_size



# Copyright (C) 2021  UAVCAN Consortium  <uavcan.org>
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

# pylint: disable=wrong-import-position

"""
Implementation of the UAVCAN register interface as defined in the UAVCAN Specification
(section 5.3 *Application-layer functions*).
"""

import uavcan.primitive
import uavcan.primitive.array

# import X as Y is not an accepted form; see https://github.com/python/mypy/issues/11706
Empty = uavcan.primitive.Empty_1
String = uavcan.primitive.String_1
Unstructured = uavcan.primitive.Unstructured_1
Bit = uavcan.primitive.array.Bit_1
Integer64 = uavcan.primitive.array.Integer64_1
Integer32 = uavcan.primitive.array.Integer32_1
Integer16 = uavcan.primitive.array.Integer16_1
Integer8 = uavcan.primitive.array.Integer8_1
Natural64 = uavcan.primitive.array.Natural64_1
Natural32 = uavcan.primitive.array.Natural32_1
Natural16 = uavcan.primitive.array.Natural16_1
Natural8 = uavcan.primitive.array.Natural8_1
Real64 = uavcan.primitive.array.Real64_1
Real32 = uavcan.primitive.array.Real32_1
Real16 = uavcan.primitive.array.Real16_1

from ._value import Value as Value
from ._value import ValueProxy as ValueProxy
from ._value import RelaxedValue as RelaxedValue
from ._value import ValueConversionError as ValueConversionError

from . import backend as backend

from ._registry import Registry as Registry
from ._registry import ValueProxyWithFlags as ValueProxyWithFlags
from ._registry import MissingRegisterError as MissingRegisterError


def get_environment_variable_name(register_name: str) -> str:
    """
    Convert the name of the register to the name of the environment variable that assigns it.

    >>> get_environment_variable_name("m.motor.inductance_dq")
    'M__MOTOR__INDUCTANCE_DQ'
    """
    return register_name.upper().replace(".", "__")



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import typing
import logging
import asyncio
import pyuavcan.util
import pyuavcan.dsdl
import pyuavcan.transport
from ._base import MessagePort, OutgoingTransferIDCounter, MessageClass, Closable
from ._base import DEFAULT_PRIORITY, PortFinalizer
from ._error import PortClosedError


_logger = logging.getLogger(__name__)


class Publisher(MessagePort[MessageClass]):
    """
    A task should request its own independent publisher instance from the presentation layer controller.
    Do not share the same publisher instance across different tasks. This class implements the RAII pattern.

    Implementation info: all publishers sharing the same session specifier (i.e., subject-ID) also share the same
    underlying implementation object containing the transport session which is reference counted and destroyed
    automatically when the last publisher with that session specifier is closed;
    the user code cannot access it and generally shouldn't care.
    None of the settings of a publisher instance, such as send timeout or priority, can affect other publishers;
    this does not apply to the transfer-ID counter objects though because they are transport-layer entities
    and therefore are shared per session specifier.
    """

    DEFAULT_SEND_TIMEOUT = 1.0
    """
    Default value for :attr:`send_timeout`. The value is an implementation detail, not required by Specification.
    """

    def __init__(self, impl: PublisherImpl[MessageClass]):
        """
        Do not call this directly! Use :meth:`Presentation.make_publisher`.
        """
        self._maybe_impl: typing.Optional[PublisherImpl[MessageClass]] = impl
        impl.register_proxy()  # Register ASAP to ensure correct finalization.

        self._dtype = impl.dtype  # Permit usage after close()
        self._transport_session = impl.transport_session  # Same
        self._transfer_id_counter = impl.transfer_id_counter  # Same
        self._priority: pyuavcan.transport.Priority = DEFAULT_PRIORITY
        self._send_timeout = self.DEFAULT_SEND_TIMEOUT

    @property
    def dtype(self) -> typing.Type[MessageClass]:
        return self._dtype

    @property
    def transport_session(self) -> pyuavcan.transport.OutputSession:
        return self._transport_session

    @property
    def transfer_id_counter(self) -> OutgoingTransferIDCounter:
        """
        Allows the caller to reach the transfer-ID counter object of this session (shared per session specifier).
        This may be useful in certain special cases such as publication of time synchronization messages,
        where it may be necessary to override the transfer-ID manually.
        """
        return self._transfer_id_counter

    @property
    def priority(self) -> pyuavcan.transport.Priority:
        """
        The priority level used for transfers published via this instance.
        This parameter is configured separately per proxy instance; i.e., it is not shared across different publisher
        instances under the same session specifier.
        """
        return self._priority

    @priority.setter
    def priority(self, value: pyuavcan.transport.Priority) -> None:
        assert value in pyuavcan.transport.Priority
        self._priority = value

    @property
    def send_timeout(self) -> float:
        """
        Every outgoing transfer initiated via this proxy instance will have to be sent in this amount of time.
        If the time is exceeded, the attempt is aborted and False is returned. Read the transport layer docs for
        an in-depth information on send timeout handling.
        The default is :attr:`DEFAULT_SEND_TIMEOUT`.
        The publication logic is roughly as follows::

            return transport_session.send(message_transfer, self.loop.time() + self.send_timeout)
        """
        return self._send_timeout

    @send_timeout.setter
    def send_timeout(self, value: float) -> None:
        value = float(value)
        if 0 < value < float("+inf"):
            self._send_timeout = value
        else:
            raise ValueError(f"Invalid send timeout value: {value}")

    async def publish(self, message: MessageClass) -> bool:
        """
        Serializes and publishes the message object at the priority level selected earlier.
        Should not be used simultaneously with :meth:`publish_soon` because that makes the message ordering undefined.
        Returns False if the publication could not be completed in :attr:`send_timeout`, True otherwise.
        """
        if self._maybe_impl is None:
            raise PortClosedError(repr(self))
        loop = asyncio.get_running_loop()
        return await self._maybe_impl.publish(message, self._priority, loop.time() + self._send_timeout)

    def publish_soon(self, message: MessageClass) -> None:
        """
        Serializes and publishes the message object at the priority level selected earlier.
        Does so without blocking (observe that this method is not async).
        Should not be used simultaneously with :meth:`publish` because that makes the message ordering undefined.
        The send timeout is still in effect here -- if the operation cannot complete in the selected time,
        send will be cancelled and a low-severity log message will be emitted.
        """
        if self._maybe_impl is None:  # Detect errors as early as possible, do not wait for the task to start.
            raise PortClosedError(repr(self))

        async def executor() -> None:
            try:
                if not await self.publish(message):
                    _logger.info("%s send timeout", self)
            except Exception as ex:
                _logger.exception("%s deferred publication has failed: %s", self, ex)

        asyncio.ensure_future(executor())

    def close(self) -> None:
        impl, self._maybe_impl = self._maybe_impl, None
        if impl is not None:
            impl.remove_proxy()

    def __del__(self) -> None:
        if self._maybe_impl is not None:
            # https://docs.python.org/3/reference/datamodel.html#object.__del__
            # DO NOT invoke logging from the finalizer because it may resurrect the object!
            # Once it is resurrected, we may run into resource management issue if __del__() is invoked again.
            # Whether it is invoked the second time is an implementation detail.
            self._maybe_impl.remove_proxy()
            self._maybe_impl = None


class PublisherImpl(Closable, typing.Generic[MessageClass]):
    """
    The publisher implementation. There is at most one such implementation per session specifier. It may be shared
    across multiple users with the help of the proxy class. When the last proxy is closed or garbage collected,
    the implementation will also be closed and removed. This is not a part of the library API.
    """

    def __init__(
        self,
        dtype: typing.Type[MessageClass],
        transport_session: pyuavcan.transport.OutputSession,
        transfer_id_counter: OutgoingTransferIDCounter,
        finalizer: PortFinalizer,
    ):
        self.dtype = dtype
        self.transport_session = transport_session
        self.transfer_id_counter = transfer_id_counter
        self._maybe_finalizer: typing.Optional[PortFinalizer] = finalizer
        self._lock = asyncio.Lock()
        self._proxy_count = 0

    async def publish(
        self, message: MessageClass, priority: pyuavcan.transport.Priority, monotonic_deadline: float
    ) -> bool:
        if not isinstance(message, self.dtype):
            raise TypeError(f"Expected a message object of type {self.dtype}, found this: {message}")

        async with self._lock:
            if self._is_closed:
                raise PortClosedError(repr(self))
            timestamp = pyuavcan.transport.Timestamp.now()
            fragmented_payload = list(pyuavcan.dsdl.serialize(message))
            transfer = pyuavcan.transport.Transfer(
                timestamp=timestamp,
                priority=priority,
                transfer_id=self.transfer_id_counter.get_then_increment(),
                fragmented_payload=fragmented_payload,
            )
            return await self.transport_session.send(transfer, monotonic_deadline)

    def register_proxy(self) -> None:
        self._proxy_count += 1
        _logger.debug("%s got a new proxy, new count %s", self, self._proxy_count)
        assert not self._is_closed, "Internal protocol violation"
        assert self._proxy_count >= 1

    def remove_proxy(self) -> None:
        self._proxy_count -= 1
        _logger.debug("%s has lost a proxy, new count %s", self, self._proxy_count)
        if self._proxy_count <= 0:
            self.close()  # RAII auto-close
        assert self._proxy_count >= 0

    @property
    def proxy_count(self) -> int:
        """Testing facilitation."""
        assert self._proxy_count >= 0
        return self._proxy_count

    def close(self) -> None:
        if self._maybe_finalizer is not None:
            self._maybe_finalizer([self.transport_session])
            self._maybe_finalizer = None

    @property
    def _is_closed(self) -> bool:
        return self._maybe_finalizer is None

    def __repr__(self) -> str:
        return pyuavcan.util.repr_attributes_noexcept(
            self,
            dtype=str(pyuavcan.dsdl.get_model(self.dtype)),
            transport_session=self.transport_session,
            proxy_count=self._proxy_count,
        )



#!/usr/bin/env python3
#
# Copyright (C) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import setuptools

setuptools.setup()



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import enum
import typing
import dataclasses
import pyuavcan.util
from ._timestamp import Timestamp


FragmentedPayload = typing.Sequence[memoryview]
"""
Transfer payload is allowed to be segmented to facilitate zero-copy implementations.
The format of the memoryview object should be 'B'.
We're using Sequence and not Iterable to permit sharing across multiple consumers.
"""


class Priority(enum.IntEnum):
    """
    Transfer priority enumeration follows the recommended names provided in the UAVCAN specification.
    We use integers here in order to allow usage of static lookup tables for conversion into transport-specific
    priority values. The particular integer values used here may be meaningless for some transports.
    """

    EXCEPTIONAL = 0
    IMMEDIATE = 1
    FAST = 2
    HIGH = 3
    NOMINAL = 4
    LOW = 5
    SLOW = 6
    OPTIONAL = 7


@dataclasses.dataclass(frozen=True)
class Transfer:
    """
    UAVCAN transfer representation.
    """

    timestamp: Timestamp
    """
    For output (tx) transfers this field contains the transfer creation timestamp.
    For input (rx) transfers this field contains the first frame reception timestamp.
    """

    priority: Priority
    """
    See :class:`Priority`.
    """

    transfer_id: int
    """
    When transmitting, the appropriate modulus will be computed by the transport automatically.
    Higher layers shall use monotonically increasing transfer-ID counters.
    """

    fragmented_payload: FragmentedPayload
    """
    See :class:`FragmentedPayload`. This is the serialized application-level payload.
    Fragmentation may be completely arbitrary.
    Received transfers usually have it fragmented such that one fragment corresponds to one received frame.
    Outgoing transfers usually fragment it according to the structure of the serialized data object.
    The purpose of fragmentation is to eliminate unnecessary data copying within the protocol stack.
    :func:`pyuavcan.transport.commons.refragment` is designed to facilitate regrouping when sending a transfer.
    """

    def __repr__(self) -> str:
        fragmented_payload = "+".join(f"{len(x)}B" for x in self.fragmented_payload)
        kwargs = {f.name: getattr(self, f.name) for f in dataclasses.fields(self)}
        kwargs["priority"] = self.priority.name
        kwargs["fragmented_payload"] = f"[{fragmented_payload}]"
        del kwargs["timestamp"]
        return pyuavcan.util.repr_attributes(self, str(self.timestamp), **kwargs)


@dataclasses.dataclass(frozen=True, repr=False)
class TransferFrom(Transfer):
    """
    Specialization for received transfers.
    """

    source_node_id: typing.Optional[int]
    """
    None indicates anonymous transfers.
    """



# Copyright (c) 2020 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

"""
Plug-and-play node-ID allocation logic. See the class documentation for usage info.

Remember that a network that contains static nodes alongside PnP nodes may encounter node-ID conflicts
when a static node appears online after its node-ID is already granted to a PnP node.
To avoid this, the Specification recommends that PnP nodes and static nodes are not to be mixed on the same network
(excepting the allocators themselves -- they are always static, naturally).
"""

from __future__ import annotations
import abc
from typing import Optional, Union, Any
import random
import asyncio
import pathlib
import logging
import sqlite3
import uavcan.node
from uavcan.pnp import NodeIDAllocationData_1 as NodeIDAllocationData_1
from uavcan.pnp import NodeIDAllocationData_2 as NodeIDAllocationData_2
import pyuavcan
import pyuavcan.application

# import X as Y is not an accepted form; see https://github.com/python/mypy/issues/11706
ID = uavcan.node.ID_1

_PSEUDO_UNIQUE_ID_MASK = (
    2 ** pyuavcan.dsdl.get_model(NodeIDAllocationData_1)["unique_id_hash"].data_type.bit_length_set.max - 1
)

_NODE_ID_MASK = 2 ** pyuavcan.dsdl.get_model(ID)["value"].data_type.bit_length_set.max - 1

_UNIQUE_ID_SIZE_BYTES = pyuavcan.application.NodeInfo().unique_id.size

_NUM_RESERVED_TOP_NODE_IDS = 2

_DB_DEFAULT_LOCATION = ":memory:"
_DB_TIMEOUT = 1.0


_logger = logging.getLogger(__name__)


class Allocatee:
    """
    Plug-and-play node-ID protocol client.

    This class represents a node that requires an allocated node-ID.
    Once started, the client will keep issuing node-ID allocation requests until either a node-ID is granted
    or until the node-ID of the specified transport instance ceases to be anonymous
    (that could happen if the transport is re-configured by the application locally).
    The status (whether the allocation is finished or still in progress) is to be queried periodically
    via :meth:`get_result`.

    Uses v1 allocation messages if the transport MTU is small (like if the transport is Classic CAN).
    Switches between v1 and v2 as necessary on the fly if the transport is reconfigured at runtime.

    Unlike other application-layer function implementations, this class takes a transport instance directly
    instead of a node because it is expected to be used before the node object is constructed.
    """

    DEFAULT_PRIORITY = pyuavcan.transport.Priority.SLOW

    _MTU_THRESHOLD = pyuavcan.dsdl.get_model(NodeIDAllocationData_2).bit_length_set.max // 8

    def __init__(
        self,
        transport_or_presentation: Union[pyuavcan.transport.Transport, pyuavcan.presentation.Presentation],
        local_unique_id: bytes,
        preferred_node_id: Optional[int] = None,
    ):
        """
        :param transport_or_presentation:
            The transport to run the allocation client on, or the presentation instance constructed on it.
            If the transport is not anonymous (i.e., a node-ID is already set),
            the allocatee will simply return the existing node-ID and do nothing.

        :param local_unique_id:
            The 128-bit globally unique-ID of the local node; the same value is also contained
            in ``uavcan.node.GetInfo.Response``.
            Beware that random generation of the unique-ID at every launch is a bad idea because it will
            exhaust the allocation table quickly.
            Refer to the Specification for details.

        :param preferred_node_id:
            If the application prefers to obtain a particular node-ID, it can specify it here.
            If provided, the PnP allocator will try to find a node-ID that is close to the stated preference.
            If not provided, the PnP allocator will pick a node-ID at its own discretion.
        """
        if isinstance(transport_or_presentation, pyuavcan.transport.Transport):
            self._transport = transport_or_presentation
            self._presentation = pyuavcan.presentation.Presentation(self._transport)
        elif isinstance(transport_or_presentation, pyuavcan.presentation.Presentation):
            self._transport = transport_or_presentation.transport
            self._presentation = transport_or_presentation
        else:  # pragma: no cover
            raise TypeError(f"Expected transport or presentation controller, found {type(transport_or_presentation)}")

        self._local_unique_id = local_unique_id
        self._preferred_node_id = int(preferred_node_id if preferred_node_id is not None else _NODE_ID_MASK)
        if not isinstance(self._local_unique_id, bytes) or len(self._local_unique_id) != _UNIQUE_ID_SIZE_BYTES:
            raise ValueError(f"Invalid unique-ID: {self._local_unique_id!r}")
        if not (0 <= self._preferred_node_id <= _NODE_ID_MASK):
            raise ValueError(f"Invalid preferred node-ID: {self._preferred_node_id}")

        self._result: Optional[int] = None
        self._sub_1 = self._presentation.make_subscriber_with_fixed_subject_id(NodeIDAllocationData_1)
        self._sub_2 = self._presentation.make_subscriber_with_fixed_subject_id(NodeIDAllocationData_2)
        self._pub: Union[
            None,
            pyuavcan.presentation.Publisher[NodeIDAllocationData_1],
            pyuavcan.presentation.Publisher[NodeIDAllocationData_2],
        ] = None
        self._timer: Optional[asyncio.TimerHandle] = None

        self._sub_1.receive_in_background(self._on_response)
        self._sub_2.receive_in_background(self._on_response)
        self._restart_timer()

    @property
    def presentation(self) -> pyuavcan.presentation.Presentation:
        return self._presentation

    def get_result(self) -> Optional[int]:
        """
        None if the allocation is still in progress. If the allocation is finished, this is the allocated node-ID.
        """
        res = self.presentation.transport.local_node_id
        return res if res is not None else self._result

    def close(self) -> None:
        """
        Stop the allocation process. The allocatee automatically closes itself shortly after the allocation is finished,
        so it's not necessary to invoke this method after a successful allocation.
        **The underlying transport is NOT closed.** The method is idempotent.
        """
        if self._timer is not None:
            self._timer.cancel()
            self._timer = None
        self._sub_1.close()
        self._sub_2.close()
        if self._pub is not None:
            self._pub.close()
            self._pub = None

    def _on_timer(self) -> None:
        self._restart_timer()
        if self.get_result() is not None:
            self.close()
            return

        msg: Any = None
        try:
            if self.presentation.transport.protocol_parameters.mtu > self._MTU_THRESHOLD:
                msg = NodeIDAllocationData_2(node_id=ID(self._preferred_node_id), unique_id=self._local_unique_id)
            else:
                msg = NodeIDAllocationData_1(unique_id_hash=_make_pseudo_unique_id(self._local_unique_id))

            if self._pub is None or self._pub.dtype != type(msg):
                if self._pub is not None:
                    self._pub.close()
                self._pub = self.presentation.make_publisher_with_fixed_subject_id(type(msg))
                self._pub.priority = self.DEFAULT_PRIORITY

            _logger.debug("Publishing allocation request %s", msg)
            self._pub.publish_soon(msg)
        except Exception as ex:
            _logger.exception("Could not send allocation request %s: %s", msg, ex)

    def _restart_timer(self) -> None:
        t_request = random.random()
        self._timer = asyncio.get_event_loop().call_later(t_request, self._on_timer)

    async def _on_response(
        self, msg: Union[NodeIDAllocationData_1, NodeIDAllocationData_2], meta: pyuavcan.transport.TransferFrom
    ) -> None:
        if self.get_result() is not None:  # Allocation already done, nothing else to do.
            return

        if meta.source_node_id is None:  # Another request, ignore.
            return

        allocated: Optional[int] = None
        if isinstance(msg, NodeIDAllocationData_1):
            if msg.unique_id_hash == _make_pseudo_unique_id(self._local_unique_id) and len(msg.allocated_node_id) > 0:
                allocated = msg.allocated_node_id[0].value
        elif isinstance(msg, NodeIDAllocationData_2):
            if msg.unique_id.tobytes() == self._local_unique_id:
                allocated = msg.node_id.value
        else:
            assert False, "Internal logic error"

        if allocated is None:
            return  # UID mismatch.

        assert isinstance(allocated, int)
        protocol_params = self.presentation.transport.protocol_parameters
        max_node_id = min(protocol_params.max_nodes - 1, _NODE_ID_MASK)
        if not (0 <= allocated <= max_node_id):
            _logger.warning(
                "Allocated node-ID %s ignored because it is incompatible with the transport: %s",
                allocated,
                protocol_params,
            )
            return

        _logger.info("Plug-and-play allocation done: got node-ID %s from server %s", allocated, meta.source_node_id)
        self._result = allocated


class Allocator:
    """
    An abstract PnP allocator interface. See derived classes.

    If an existing allocation table is reused with a least capable transport where the maximum node-ID is smaller,
    the allocator may create redundant allocations in order to avoid granting node-ID values that exceed the valid
    node-ID range for the transport.
    """

    DEFAULT_PUBLICATION_TIMEOUT = 5.0
    """
    The allocation message publication timeout is chosen to be large because the data is constant
    (does not lose relevance over time) and the priority level is usually low.
    """

    @abc.abstractmethod
    def register_node(self, node_id: int, unique_id: Optional[bytes]) -> None:
        """
        This method shall be invoked whenever a new node appears online and/or whenever its unique-ID is obtained.
        The recommended usage pattern is to subscribe to the update events from
        :class:`pyuavcan.application.node_tracker.NodeTracker`, where the necessary update logic is already implemented.
        """
        raise NotImplementedError


class CentralizedAllocator(Allocator):
    """
    The centralized plug-and-play node-ID allocator. See Specification for details.
    """

    def __init__(
        self,
        node: pyuavcan.application.Node,
        database_file: Optional[Union[str, pathlib.Path]] = None,
    ):
        """
        :param node:
            The node instance to run the allocator on.
            The 128-bit globally unique-ID of the local node will be sourced from this instance.
            Refer to the Specification for details.

        :param database_file:
            If provided, shall specify the path to the database file containing an allocation table.
            If the file does not exist, it will be automatically created. If None (default), the allocation table
            will be created in memory (therefore the allocation data will be lost after the instance is disposed).
        """
        self._node = node
        local_node_id = self.node.id
        if local_node_id is None:
            raise ValueError("The allocator cannot run on an anonymous node")
        # The database is initialized with ``check_same_thread=False`` to enable delegating its initialization
        # to a thread pool from an async context. This is important for this library because if one needs to
        # initialize a new instance from an async function, running the initialization directly may be unacceptable
        # due to its blocking behavior, so one is likely to rely on :meth:`asyncio.loop.run_in_executor`.
        # The executor will initialize the instance in a worker thread and then hand it over to the main thread,
        # which is perfectly safe, but it would trigger a false error from the SQLite engine complaining about
        # the possibility of concurrency-related bugs.
        self._alloc = _AllocationTable(
            sqlite3.connect(str(database_file or _DB_DEFAULT_LOCATION), timeout=_DB_TIMEOUT, check_same_thread=False)
        )
        self._alloc.register(local_node_id, self.node.info.unique_id.tobytes())
        self._sub1 = self.node.make_subscriber(NodeIDAllocationData_1)
        self._sub2 = self.node.make_subscriber(NodeIDAllocationData_2)
        self._pub1 = self.node.make_publisher(NodeIDAllocationData_1)
        self._pub2 = self.node.make_publisher(NodeIDAllocationData_2)
        self._pub1.send_timeout = self.DEFAULT_PUBLICATION_TIMEOUT
        self._pub2.send_timeout = self.DEFAULT_PUBLICATION_TIMEOUT

        def start() -> None:
            _logger.debug("Centralized allocator starting with the following allocation table:\n%s", self._alloc)
            self._sub1.receive_in_background(self._on_message)
            self._sub2.receive_in_background(self._on_message)

        def close() -> None:
            for port in [self._sub1, self._sub2, self._pub1, self._pub2]:
                assert isinstance(port, pyuavcan.presentation.Port)
                port.close()
            self._alloc.close()

        node.add_lifetime_hooks(start, close)

    @property
    def node(self) -> pyuavcan.application.Node:
        return self._node

    def register_node(self, node_id: int, unique_id: Optional[bytes]) -> None:
        self._alloc.register(node_id, unique_id)

    async def _on_message(
        self, msg: Union[NodeIDAllocationData_1, NodeIDAllocationData_2], meta: pyuavcan.transport.TransferFrom
    ) -> None:
        if meta.source_node_id is not None:
            _logger.error(  # pylint: disable=logging-fstring-interpolation
                f"Invalid network configuration: another node-ID allocator detected at node-ID {meta.source_node_id}. "
                f"There shall be exactly one allocator on the network. If modular redundancy is desired, "
                f"use a distributed allocator (currently, a centralized allocator is running). "
                f"The detected allocation response message is {msg} with metadata {meta}."
            )
            return

        _logger.debug("Received allocation request %s with metadata %s", msg, meta)
        max_node_id = self.node.presentation.transport.protocol_parameters.max_nodes - 1 - _NUM_RESERVED_TOP_NODE_IDS
        assert max_node_id > 0

        if isinstance(msg, NodeIDAllocationData_1):
            allocated = self._alloc.allocate(max_node_id, max_node_id, pseudo_unique_id=msg.unique_id_hash)
            if allocated is not None:
                self._respond_v1(meta.priority, msg.unique_id_hash, allocated)
                return
        elif isinstance(msg, NodeIDAllocationData_2):
            uid = msg.unique_id.tobytes()
            allocated = self._alloc.allocate(msg.node_id.value, max_node_id, unique_id=uid)
            if allocated is not None:
                self._respond_v2(meta.priority, uid, allocated)
                return
        else:
            assert False, "Internal logic error"
        _logger.warning("Allocation table is full, ignoring request %s with %s. Please purge the table.", msg, meta)

    def _respond_v1(self, priority: pyuavcan.transport.Priority, unique_id_hash: int, allocated_node_id: int) -> None:
        msg = NodeIDAllocationData_1(unique_id_hash=unique_id_hash, allocated_node_id=[ID(allocated_node_id)])
        _logger.info("Publishing allocation response v1: %s", msg)
        self._pub1.priority = priority
        self._pub1.publish_soon(msg)

    def _respond_v2(self, priority: pyuavcan.transport.Priority, unique_id: bytes, allocated_node_id: int) -> None:
        msg = NodeIDAllocationData_2(
            node_id=ID(allocated_node_id),
            unique_id=unique_id,
        )
        _logger.info("Publishing allocation response v2: %s", msg)
        self._pub2.priority = priority
        self._pub2.publish_soon(msg)


class DistributedAllocator(Allocator):
    """
    This class is a placeholder. The implementation is missing (could use help here).
    The implementation can be based on the existing distributed allocator from Libuavcan v0,
    although the new PnP protocol is much simpler because it lacks multi-stage exchanges.
    """

    def __init__(self, node: pyuavcan.application.Node):
        assert node
        raise NotImplementedError((self.__doc__ or "").strip())

    def register_node(self, node_id: int, unique_id: Optional[bytes]) -> None:
        raise NotImplementedError


class _AllocationTable:
    _SCHEMA = """
    create table if not exists `allocation` (
        `node_id`          int not null unique check(node_id >= 0),
        `unique_id_hex`    varchar(32) not null,  -- all zeros if unique-ID is unknown.
        `pseudo_unique_id` bigint not null check(pseudo_unique_id >= 0), -- 48 LSB of CRC64WE(unique-ID); v1 compat.
        `ts`               time not null default current_timestamp,
        primary key(node_id)
    );
    """

    def __init__(self, db_connection: sqlite3.Connection):
        self._con = db_connection
        self._con.execute(self._SCHEMA)
        self._con.commit()

    def register(self, node_id: int, unique_id: Optional[bytes]) -> None:
        unique_id_defined = unique_id is not None
        if unique_id is None:
            unique_id = bytes(_UNIQUE_ID_SIZE_BYTES)
        if not isinstance(unique_id, bytes) or len(unique_id) != _UNIQUE_ID_SIZE_BYTES:
            raise ValueError(f"Invalid unique-ID: {unique_id!r}")
        if not isinstance(node_id, int) or not (0 <= node_id <= _NODE_ID_MASK):
            raise ValueError(f"Invalid node-ID: {node_id!r}")

        def execute() -> None:
            assert isinstance(unique_id, bytes)
            self._con.execute(
                "insert or replace into allocation (node_id, unique_id_hex, pseudo_unique_id) values (?, ?, ?);",
                (node_id, unique_id.hex(), _make_pseudo_unique_id(unique_id)),
            )
            self._con.commit()

        res = self._con.execute("select unique_id_hex from allocation where node_id = ?", (node_id,)).fetchone()
        existing_uid = bytes.fromhex(res[0]) if res is not None else None
        if existing_uid is None:
            _logger.debug("Original node registration: NID % 5d, UID %s", node_id, unique_id.hex())
            execute()
        elif unique_id_defined and existing_uid != unique_id:
            _logger.debug(
                "Updated node registration:  NID % 5d, UID %s -> %s", node_id, existing_uid.hex(), unique_id.hex()
            )
            execute()

    def allocate(
        self,
        preferred_node_id: int,
        max_node_id: int,
        unique_id: Optional[bytes] = None,
        pseudo_unique_id: Optional[int] = None,
    ) -> Optional[int]:
        use_unique_id = unique_id is not None
        preferred_node_id = min(max(preferred_node_id, 0), max_node_id)
        _logger.debug(
            "Table alloc request: preferred_node_id=%s, max_node_id=%s, unique_id=%s, pseudo_unique_id=%s",
            preferred_node_id,
            max_node_id,
            unique_id.hex() if unique_id else None,
            pseudo_unique_id,
        )
        if unique_id is None:
            unique_id = bytes(_UNIQUE_ID_SIZE_BYTES)
        if pseudo_unique_id is None:
            pseudo_unique_id = _make_pseudo_unique_id(unique_id)
        assert isinstance(unique_id, bytes) and len(unique_id) == _UNIQUE_ID_SIZE_BYTES
        assert isinstance(pseudo_unique_id, int) and (0 <= pseudo_unique_id <= _PSEUDO_UNIQUE_ID_MASK)

        # Check if there is an existing allocation for this UID. If there are multiple matches, pick the newest.
        # Ignore existing allocations where the node-ID exceeds the maximum in case we're reusing an existing
        # allocation table with a less capable transport.
        if use_unique_id:
            res = self._con.execute(
                "select node_id from allocation where unique_id_hex = ? and node_id <= ? order by ts desc limit 1",
                (unique_id.hex(), max_node_id),
            ).fetchone()
        else:
            res = self._con.execute(
                "select node_id from allocation where pseudo_unique_id = ? and node_id <= ? order by ts desc limit 1",
                (pseudo_unique_id, max_node_id),
            ).fetchone()
        if res is not None:
            candidate = int(res[0])
            assert 0 <= candidate <= max_node_id, "Internal logic error"
            _logger.debug(
                "Serving existing allocation: NID %s, (pseudo-)UID %s",
                candidate,
                unique_id.hex() if use_unique_id else hex(pseudo_unique_id),
            )
            return candidate

        # Do a new allocation. Consider re-implementing this in pure SQL -- should be possible with SQLite.
        result: Optional[int] = None
        candidate = preferred_node_id
        while result is None and candidate <= max_node_id:
            if self._try_allocate(candidate, unique_id, pseudo_unique_id):
                result = candidate
            candidate += 1
        candidate = preferred_node_id
        while result is None and candidate >= 0:
            if self._try_allocate(candidate, unique_id, pseudo_unique_id):
                result = candidate
            candidate -= 1

        # Final report.
        if result is not None:
            _logger.debug(
                "New allocation: allocated NID %s, (pseudo-)UID %s, preferred NID %s",
                result,
                unique_id.hex() if use_unique_id else hex(pseudo_unique_id),
                preferred_node_id,
            )
        return result

    def close(self) -> None:
        self._con.close()

    def _try_allocate(self, node_id: int, unique_id: bytes, pseudo_unique_id: int) -> bool:
        try:
            self._con.execute(
                "insert into allocation (node_id, unique_id_hex, pseudo_unique_id) values (?, ?, ?);",
                (node_id, unique_id.hex(), pseudo_unique_id),
            )
            self._con.commit()
        except sqlite3.IntegrityError:  # Such entry already exists.
            return False
        return True

    def __str__(self) -> str:
        """Displays the table as a multi-line string in TSV format with one header line."""
        lines = ["Node-ID\t" + "Unique-ID/hash (hex)".ljust(32 + 1 + 12) + "\tUpdate timestamp"]
        for nid, uid_hex, pseudo, ts in self._con.execute(
            "select node_id, unique_id_hex, pseudo_unique_id, ts from allocation order by ts desc"
        ).fetchall():
            lines.append(f"{nid: 5d}  \t{uid_hex:32s}/{pseudo:012x}\t{ts}")
        return "\n".join(lines) + "\n"


def _make_pseudo_unique_id(unique_id: bytes) -> int:
    """
    The recommended mapping function from unique-ID to pseudo unique-ID.
    """
    from pyuavcan.transport.commons.crc import CRC64WE

    assert isinstance(unique_id, bytes) and len(unique_id) == _UNIQUE_ID_SIZE_BYTES
    return int(CRC64WE.new(unique_id).value & _PSEUDO_UNIQUE_ID_MASK)



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

"""
UAVCAN/Serial transport overview
++++++++++++++++++++++++++++++++

The UAVCAN/Serial transport is experimental and is not yet part of the UAVCAN specification.
Future revisions may break wire compatibility until the transport is formally specified.
Context: https://forum.uavcan.org/t/alternative-transport-protocols/324, also see the discussion at
https://forum.uavcan.org/t/yukon-design-megathread/390/115?u=pavel.kirienko.

The UAVCAN/Serial transport is designed for OSI L1 byte-level duplex serial links and tunnels:

- UART, RS-422/485/232 (duplex); the recommended rates are: 115200 bps, 921600 bps, 3 Mbps, 10 Mbps, 100 Mbps.
- USB CDC ACM.
- TCP/IP encapsulation.

It may also be suited for raw transport log storage, because one-dimensional flat binary files are structurally
similar to serial byte-level links.

This transport module contains no media sublayers because the media abstraction
is handled directly by the `PySerial <https://pypi.org/project/pyserial>`_
library and the underlying operating system.

The serial transport supports all transfer categories:

+--------------------+--------------------------+---------------------------+
| Supported transfers| Unicast                  | Broadcast                 |
+====================+==========================+===========================+
|**Message**         | Yes, non-spec extension  | Yes                       |
+--------------------+--------------------------+---------------------------+
|**Service**         | Yes                      | Banned by Specification   |
+--------------------+--------------------------+---------------------------+


Protocol definition
+++++++++++++++++++

The packet header is defined as follows (byte/bit ordering in this definition follow the DSDL specification:
least significant first)::

    uint8   version              # Always zero. Discard the frame if not.
    uint8   priority             # 0 = highest, 7 = lowest; the rest are unused.
    uint16  source node ID       # 0xFFFF = anonymous.
    uint16  destination node ID  # 0xFFFF = broadcast.
    uint16  data specifier

    void64                       # Reserved; later may be leveraged for runtime type identification.
    uint64  transfer-ID

    uint32  frame index EOT      # MSB set if last frame of the transfer; i.e., 0x8000_0000 if single-frame transfer.
    uint32  header CRC           # CRC-32C (Castagnoli) of the header (all fields above).

For message frames, the data specifier field contains the subject-ID value,
so that the most significant bit is always cleared.
For service frames, the most significant bit (15th) is always set,
and the second-to-most-significant bit (14th) is set for response transfers only;
the remaining 14 least significant bits contain the service-ID value.

Total header size: 32 bytes (256 bits).

The header is prepended before the frame payload; the resulting structure is
encoded into its serialized form using the following packet format:

+-------------------------+--------------+---------------+--------------------------------+-------------------------+
| Frame delimiter **0x00**|Escaped header|Escaped payload| Escaped CRC-32C of the payload | Frame delimiter **0x00**|
+=========================+==============+===============+================================+=========================+
| 1 byte                  | 32 bytes     | >=0 bytes     | 4 bytes                        | 1 byte                  |
+-------------------------+--------------+---------------+--------------------------------+-------------------------+
| Single-byte frame       |                              | Four bytes long, little-endian | Same frame delimiter as |
| delimiter **0x00**.     |                              | byte order; The CRC is         | at the start.           |
| Begins a new frame and  |                              | computed over the unescaped    | Terminates the current  |
| possibly terminates the |                              | (i.e., original form) payload, | frame and possibly      |
| previous frame.         |                              | not including the header       | begins the next frame.  |
|                         |                              | (because the header has a      |                         |
|                         |                              | dedicated CRC).                |                         |
|                         +------------------------------+--------------------------------+                         |
|                         | This part is escaped using COBS alorithm by Chesire and Baker |                         |
|                         | http://www.stuartcheshire.org/papers/COBSforToN.pdf.          |                         |
|                         | A frame delimiter (0) is guaranteed to never occur here.      |                         |
+-------------------------+------------------------------+--------------------------------+-------------------------+

The frame encoding overhead is 1 byte in every 254 bytes of the header+payload+CRC, which is about ~0.4%.
There is a somewhat relevant discussion at
https://forum.uavcan.org/t/uavcan-serial-issues-with-dma-friendliness-and-bandwidth-overhead/846.

The last four bytes of a multi-frame transfer payload contain the CRC32C (Castagnoli) hash of the transfer
payload in little-endian byte order.
The multi-frame transfer logic (decomposition and reassembly) is implemented in a separate
transport-agnostic module :mod:`pyuavcan.transport.commons.high_overhead_transport`.
**Despite the fact that the support for multi-frame transfers is built into this transport,
it should not be relied on and it may be removed later.**
The reason is that serial links do not have native support for framing, and as such,
it is possible to configure the MTU to be arbitrarily high to avoid multi-frame transfers completely.
The lack of multi-frame transfers simplifies implementations drastically, which is important for
deeply-embedded systems. As such, all serial transfers should be single-frame transfers.

Note that we use CRC-32C (Castagnoli) as the header/frame CRC instead of CRC-32K2 (Koopman-2)
which is superior at short data blocks offering the Hamming distance of 6 as opposed to 4.
This is because Castagnoli is superior for transfer CRC which is often sufficiently long
to flip the balance in favor of Castagnoli rather than Koopman.
We could use Koopman for the header/frame CRC and keep Castagnoli for the transfer CRC,
but such diversity is harmful because it would require implementers to keep two separate CRC tables
which may be costly in embedded applications and may deteriorate the performance of CPU caches.


Unreliable links and temporal redundancy
++++++++++++++++++++++++++++++++++++++++

The serial transport supports the deterministic data loss mitigation option,
where a transfer can be repeated several times to reduce the probability of its loss.
This feature is discussed in detail in the documentation for the UDP transport :mod:`pyuavcan.transport.udp`.


Usage
+++++

..  doctest::
    :hide:

    >>> import tests
    >>> tests.asyncio_allow_event_loop_access_from_top_level()
    >>> from tests import doctest_await

>>> import asyncio
>>> import pyuavcan
>>> import pyuavcan.transport.serial
>>> tr = pyuavcan.transport.serial.SerialTransport('loop://', local_node_id=1234, baudrate=115200)
>>> tr.local_node_id
1234
>>> tr.serial_port.baudrate
115200
>>> pm = pyuavcan.transport.PayloadMetadata(1024)
>>> ds = pyuavcan.transport.MessageDataSpecifier(2345)
>>> pub = tr.get_output_session(pyuavcan.transport.OutputSessionSpecifier(ds, None), pm)
>>> sub = tr.get_input_session(pyuavcan.transport.InputSessionSpecifier(ds, None), pm)
>>> doctest_await(pub.send(pyuavcan.transport.Transfer(pyuavcan.transport.Timestamp.now(),
...                                                    pyuavcan.transport.Priority.LOW,
...                                                    1111,
...                                                    fragmented_payload=[]),
...                        asyncio.get_event_loop().time() + 1.0))
True
>>> doctest_await(sub.receive(asyncio.get_event_loop().time() + 1.0))
TransferFrom(..., transfer_id=1111, ...)
>>> tr.close()


Tooling
+++++++

Serial data logging
~~~~~~~~~~~~~~~~~~~

The underlying PySerial library provides a convenient method of logging exchange through a serial port into a file.
To invoke this feature, embed the name of the serial port into the URI ``spy:///dev/ttyUSB0?file=dump.txt``,
where ``/dev/ttyUSB0`` is the name of the serial port, ``dump.txt`` is the name of the log file.


TCP/IP tunneling
~~~~~~~~~~~~~~~~

For testing or experimentation it is often convenient to use a virtual link instead of a real one.
The underlying PySerial library supports tunneling of raw serial data over TCP connections,
which can be leveraged for local testing without accessing any physical serial ports.
This option can be accessed by specifying the URI of the form ``socket://<address>:<port>``
instead of a real serial port name when establishing the connection.

The location specified in the URL must point to the TCP server port that will forward the data
to and from the other end of the link.
While such a server can be trivially coded manually by the developer,
it is possible to avoid the effort by relying on the TCP connection brokering mode available in
Ncat (which is a part of the `Nmap <https://nmap.org>`_ project, thanks Fyodor).

For example, one could set up the TCP broker as follows
(add ``-v`` to see what's happening; more info at https://nmap.org/ncat/guide/ncat-broker.html)
(the port number is chosen at random here)::

    ncat --broker --listen -p 50905

And then use a serial transport with ``socket://127.0.0.1:50905``
(N.B.: using ``localhost`` may significantly increase initialization latency on Windows due to slow DNS lookup).
All nodes whose transports are configured like that will be able to communicate with each other,
as if they were connected to the same bus.
Essentially, this can be seen as a virtualized RS-485 bus,
where same concerns regarding medium access coordination apply.

The location of the URI doesn't have to be local, of course --
one can use this approach to link UAVCAN nodes via conventional IP networks.

The exchange over the virtual bus can be dumped trivially for analysis::

    nc localhost 50905 > dump.bin


Inheritance diagram
+++++++++++++++++++

.. inheritance-diagram:: pyuavcan.transport.serial._serial
                         pyuavcan.transport.serial._frame
                         pyuavcan.transport.serial._session._base
                         pyuavcan.transport.serial._session._input
                         pyuavcan.transport.serial._session._output
                         pyuavcan.transport.serial._tracer
   :parts: 1
"""

from ._serial import SerialTransport as SerialTransport
from ._serial import SerialTransportStatistics as SerialTransportStatistics

from ._session import SerialSession as SerialSession
from ._session import SerialInputSession as SerialInputSession
from ._session import SerialOutputSession as SerialOutputSession
from ._session import SerialFeedback as SerialFeedback
from ._session import SerialInputSessionStatistics as SerialInputSessionStatistics

from ._frame import SerialFrame as SerialFrame

from ._tracer import SerialCapture as SerialCapture
from ._tracer import SerialTracer as SerialTracer
from ._tracer import SerialErrorTrace as SerialErrorTrace
from ._tracer import SerialOutOfBandTrace as SerialOutOfBandTrace



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import typing

import pyuavcan.transport


TransferRouter = typing.Callable[[pyuavcan.transport.Transfer, float], typing.Awaitable[bool]]


class LoopbackFeedback(pyuavcan.transport.Feedback):
    def __init__(self, transfer_timestamp: pyuavcan.transport.Timestamp):
        self._transfer_timestamp = transfer_timestamp

    @property
    def original_transfer_timestamp(self) -> pyuavcan.transport.Timestamp:
        return self._transfer_timestamp

    @property
    def first_frame_transmission_timestamp(self) -> pyuavcan.transport.Timestamp:
        return self._transfer_timestamp


class LoopbackOutputSession(pyuavcan.transport.OutputSession):
    def __init__(
        self,
        specifier: pyuavcan.transport.OutputSessionSpecifier,
        payload_metadata: pyuavcan.transport.PayloadMetadata,
        closer: typing.Callable[[], None],
        router: TransferRouter,
    ):
        self._specifier = specifier
        self._payload_metadata = payload_metadata
        self._closer = closer
        self._router = router
        self._stats = pyuavcan.transport.SessionStatistics()
        self._feedback_handler: typing.Optional[typing.Callable[[pyuavcan.transport.Feedback], None]] = None
        self._injected_exception: typing.Optional[Exception] = None
        self._should_timeout = False

    def enable_feedback(self, handler: typing.Callable[[pyuavcan.transport.Feedback], None]) -> None:
        self._feedback_handler = handler

    def disable_feedback(self) -> None:
        self._feedback_handler = None

    async def send(self, transfer: pyuavcan.transport.Transfer, monotonic_deadline: float) -> bool:
        if self._injected_exception is not None:
            raise self._injected_exception

        out = False if self._should_timeout else await self._router(transfer, monotonic_deadline)
        if out:
            self._stats.transfers += 1
            self._stats.frames += 1
            self._stats.payload_bytes += sum(map(len, transfer.fragmented_payload))
            if self._feedback_handler is not None:
                self._feedback_handler(LoopbackFeedback(transfer.timestamp))
        else:
            self._stats.drops += 1

        return out

    @property
    def specifier(self) -> pyuavcan.transport.OutputSessionSpecifier:
        return self._specifier

    @property
    def payload_metadata(self) -> pyuavcan.transport.PayloadMetadata:
        return self._payload_metadata

    def sample_statistics(self) -> pyuavcan.transport.SessionStatistics:
        return self._stats

    def close(self) -> None:
        self._injected_exception = pyuavcan.transport.ResourceClosedError(f"{self} is closed")
        self._closer()

    @property
    def exception(self) -> typing.Optional[Exception]:
        """
        This is a test rigging.
        Use this property to configure an exception object that will be raised when :func:`send` is invoked.
        Set None to remove the injected exception (None is the default value).
        Useful for testing error handling logic.
        """
        return self._injected_exception

    @exception.setter
    def exception(self, value: typing.Optional[Exception]) -> None:
        if isinstance(value, Exception) or value is None:
            self._injected_exception = value
        else:
            raise ValueError(f"Bad exception: {value}")

    @property
    def should_timeout(self) -> bool:
        return self._should_timeout

    @should_timeout.setter
    def should_timeout(self, value: bool) -> None:
        self._should_timeout = bool(value)


def _unittest_session() -> None:
    closed = False

    specifier = pyuavcan.transport.OutputSessionSpecifier(pyuavcan.transport.MessageDataSpecifier(123), 123)
    payload_metadata = pyuavcan.transport.PayloadMetadata(1234)

    def do_close() -> None:
        nonlocal closed
        closed = True

    async def do_route(_a: pyuavcan.transport.Transfer, _b: float) -> bool:
        raise NotImplementedError

    ses = LoopbackOutputSession(
        specifier=specifier, payload_metadata=payload_metadata, closer=do_close, router=do_route
    )

    assert specifier == ses.specifier
    assert payload_metadata == ses.payload_metadata

    assert not closed
    ses.close()
    assert closed

    ts = pyuavcan.transport.Timestamp.now()
    fb = LoopbackFeedback(ts)
    assert fb.first_frame_transmission_timestamp == ts
    assert fb.original_transfer_timestamp == ts



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Alex Kiselev <a.kiselev@volz-servos.com>, Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import time
import typing
import asyncio
import logging
import threading
import functools
import dataclasses
import collections
import concurrent.futures
import warnings

import can  # type: ignore
from pyuavcan.transport import Timestamp, ResourceClosedError, InvalidMediaConfigurationError
from pyuavcan.transport.can.media import Media, FilterConfiguration, Envelope, FrameFormat, DataFrame


_logger = logging.getLogger(__name__)


class PythonCANMedia(Media):
    """
    Media interface adapter for `Python-CAN <https://python-can.readthedocs.io/>`_.
    It is designed to be usable with all host platforms supported by Python-CAN (GNU/Linux, Windows, macOS).
    Please refer to the Python-CAN documentation for information about supported CAN hardware, its configuration,
    and how to install the dependencies properly.

    This media interface supports both Classic CAN and CAN FD. The selection logic is documented below.

    This implementation emulates loopback instead of requesting it from the underlying driver due to the limitations
    of Python-CAN. Same goes for timestamping.
    If accurate timestamping is desired, consider using the SocketCAN media driver instead.

    Here is a basic usage example based on the Yakut CLI tool.
    Suppose that there are two interconnected CAN bus adapters connected to the host computer:
    one SLCAN-based, the other is PCAN USB.
    Launch Yakut to listen for messages using the SLCAN adapter::

        export YAKUT_TRANSPORT='CAN(can.media.pythoncan.PythonCANMedia("slcan:/dev/ttyACM0", 1_000_000), None)'
        yakut sub 33.uavcan.si.unit.voltage.Scalar.1.0

    While the subscriber is running, publish a message to this subject::

        export YAKUT_TRANSPORT='CAN(can.media.pythoncan.PythonCANMedia("pcan:PCAN_USBBUS1", 1_000_000), 42)'
        yakut pub 33.uavcan/si/unit/voltage/Scalar_1_0 '{volt: 12}'
    """

    _MAXIMAL_TIMEOUT_SEC = 0.1

    def __init__(
        self,
        iface_name: str,
        bitrate: typing.Union[int, typing.Tuple[int, int]],
        mtu: typing.Optional[int] = None,
        *,
        loop: typing.Optional[asyncio.AbstractEventLoop] = None,
    ) -> None:
        """
        :param iface_name: Interface name consisting of Python-CAN interface module name and its channel,
            separated with a colon. Supported interfaces are documented below.
            The semantics of the channel name are described in the documentation for Python-CAN.

            - Interface ``socketcan`` is implemented by :class:`can.interfaces.socketcan.SocketcanBus`.
              The bit rate values are only used to select Classic/FD mode.
              It is not possible to configure the actual CAN bit rate using this API.
              Example: ``socketcan:vcan0``

            - Interface ``kvaser`` is implemented by :class:`can.interfaces.kvaser.canlib.KvaserBus`.
              Example: ``kvaser:0``

            - Interface ``slcan`` is implemented by :class:`can.interfaces.slcan.slcanBus`.
              Only Classic CAN is supported.
              The serial port settings are fixed at 115200-8N1.
              Example: ``slcan:COM12``

            - Interface ``pcan`` is implemented by :class:`can.interfaces.pcan.PcanBus`.
              Ensure that `PCAN-Basic <https://www.peak-system.com/PCAN-Basic.239.0.html>`_ is installed.
              Example: ``pcan:PCAN_USBBUS1``

            - Interface ``virtual`` is described in https://python-can.readthedocs.io/en/master/interfaces/virtual.html.
              The channel name should be empty.
              Example: ``virtual:``

            - Interface ``usb2can`` is described in https://python-can.readthedocs.io/en/stable/interfaces/usb2can.html.
              Example: ``usb2can:ED000100``

            - Interface ``canalystii`` is described in
              https://python-can.readthedocs.io/en/stable/interfaces/canalystii.html.
              You need to download CANalyst library for python-can package or you can install python-can by:
              ``pip3 install git+https://github.com/Cherish-Gww/python-can.git@add_canalystii_so``
              More info: https://github.com/UAVCAN/pyuavcan/issues/178#issuecomment-912497882
              Example: ``canalystii:0``

        :param bitrate: Bit rate value in bauds; either a single integer or a tuple:

            - A single integer selects Classic CAN.
            - A tuple of two selects CAN FD, where the first integer defines the arbitration (nominal) bit rate
              and the second one defines the data phase bit rate.
            - If MTU (see below) is given and is greater than 8 bytes, CAN FD is used regardless of the above.
            - An MTU of 8 bytes and a tuple of two identical bit rates selects Classic CAN.

        :param mtu: The maximum CAN data field size in bytes.
            If provided, this value must belong to :attr:`Media.VALID_MTU_SET`.
            If not provided, the default is determined as follows:

            - If `bitrate` is a single integer: classic CAN is assumed, MTU defaults to 8 bytes.
            - If `bitrate` is two integers: CAN FD is assumed, MTU defaults to 64 bytes.

        :param loop: Deprecated.

        :raises: :class:`InvalidMediaConfigurationError` if the specified media instance
            could not be constructed, the interface name is unknown,
            or if the underlying library raised a :class:`can.CanError`.

        Use virtual bus with various bit rate and FD configurations:

        >>> media = PythonCANMedia('virtual:', 500_000)
        >>> media.is_fd, media.mtu
        (False, 8)
        >>> media = PythonCANMedia('virtual:', (500_000, 2_000_000))
        >>> media.is_fd, media.mtu
        (True, 64)
        >>> media = PythonCANMedia('virtual:', 1_000_000, 16)
        >>> media.is_fd, media.mtu
        (True, 16)

        Use PCAN-USB channel 1 in FD mode with nominal bitrate 500 kbit/s, data bitrate 2 Mbit/s, MTU 64 bytes::

            PythonCANMedia('pcan:PCAN_USBBUS1', (500_000, 2_000_000))

        Use Kvaser channel 0 in classic mode with bitrate 500k::

            PythonCANMedia('kvaser:0', 500_000)

        Use CANalyst-II channel 0 in classic mode with bitrate 500k::

            PythonCANMedia('canalystii:0', 500_000)

        """
        self._conn_name = str(iface_name).split(":")
        if len(self._conn_name) != 2:
            raise InvalidMediaConfigurationError(
                f"Interface name {iface_name!r} does not match the format 'interface:channel'"
            )
        if loop:
            warnings.warn("The loop argument is deprecated", DeprecationWarning)

        single_bitrate = isinstance(bitrate, (int, float))
        bitrate = (int(bitrate), int(bitrate)) if single_bitrate else (int(bitrate[0]), int(bitrate[1]))  # type: ignore

        default_mtu = min(self.VALID_MTU_SET) if single_bitrate else 64
        self._mtu = int(mtu) if mtu is not None else default_mtu
        if self._mtu not in self.VALID_MTU_SET:
            raise InvalidMediaConfigurationError(f"Wrong MTU value: {mtu}")

        self._is_fd = (self._mtu > min(self.VALID_MTU_SET) or not single_bitrate) and not (
            self._mtu == min(self.VALID_MTU_SET) and bitrate[0] == bitrate[1]
        )

        self._closed = False
        self._maybe_thread: typing.Optional[threading.Thread] = None
        self._rx_handler: typing.Optional[Media.ReceivedFramesHandler] = None
        self._background_executor = concurrent.futures.ThreadPoolExecutor(max_workers=1)

        params: typing.Union[_FDInterfaceParameters, _ClassicInterfaceParameters]
        if self._is_fd:
            params = _FDInterfaceParameters(
                interface_name=self._conn_name[0], channel_name=self._conn_name[1], bitrate=bitrate
            )
        else:
            params = _ClassicInterfaceParameters(
                interface_name=self._conn_name[0], channel_name=self._conn_name[1], bitrate=bitrate[0]
            )
        try:
            self._bus = _CONSTRUCTORS[self._conn_name[0]](params)
        except can.CanError as ex:
            raise InvalidMediaConfigurationError(f"Could not initialize PythonCAN: {ex}") from ex
        super().__init__()

    @property
    def interface_name(self) -> str:
        return ":".join(self._conn_name)

    @property
    def mtu(self) -> int:
        return self._mtu

    @property
    def number_of_acceptance_filters(self) -> int:
        """
        The value is currently fixed at 1 for all interfaces.
        TODO: obtain the number of acceptance filters from Python-CAN.
        """
        return 1

    @property
    def is_fd(self) -> bool:
        """
        Introspection helper. The value is True if the underlying interface operates in CAN FD mode.
        """
        return self._is_fd

    def start(self, handler: Media.ReceivedFramesHandler, no_automatic_retransmission: bool) -> None:
        if self._maybe_thread is None:
            self._rx_handler = handler
            self._maybe_thread = threading.Thread(
                target=self._thread_function, args=(asyncio.get_event_loop(),), name=str(self), daemon=True
            )
            self._maybe_thread.start()
            if no_automatic_retransmission:
                _logger.info("%s non-automatic retransmission is not supported", self)
        else:
            raise RuntimeError("The RX frame handler is already set up")

    def configure_acceptance_filters(self, configuration: typing.Sequence[FilterConfiguration]) -> None:
        if self._closed:
            raise ResourceClosedError(repr(self))
        filters = []
        for f in configuration:
            d = {"can_id": f.identifier, "can_mask": f.mask}
            if f.format is not None:  # Per Python-CAN docs, if "extended" is not set, both base/ext will be accepted.
                d["extended"] = f.format == FrameFormat.EXTENDED
            filters.append(d)
        _logger.debug("%s: Acceptance filters activated: %s", self, ", ".join(map(str, configuration)))
        self._bus.set_filters(filters)

    async def send(self, frames: typing.Iterable[Envelope], monotonic_deadline: float) -> int:
        num_sent = 0
        loopback: typing.List[typing.Tuple[Timestamp, Envelope]] = []
        loop = asyncio.get_running_loop()
        for f in frames:
            if self._closed:
                raise ResourceClosedError(repr(self))
            message = can.Message(
                arbitration_id=f.frame.identifier,
                is_extended_id=(f.frame.format == FrameFormat.EXTENDED),
                data=f.frame.data,
                is_fd=self._is_fd,
            )
            try:
                await loop.run_in_executor(
                    self._background_executor,
                    functools.partial(self._bus.send, message, timeout=monotonic_deadline - loop.time()),
                )
            except (asyncio.TimeoutError, can.CanError):  # CanError is also used to report timeouts (weird).
                break
            else:
                num_sent += 1
                if f.loopback:
                    loopback.append((Timestamp.now(), f))
        if loopback:
            loop.call_soon(self._invoke_rx_handler, loopback)
        return num_sent

    def close(self) -> None:
        self._closed = True
        try:
            if self._maybe_thread is not None:
                self._maybe_thread.join(timeout=self._MAXIMAL_TIMEOUT_SEC * 10)
                self._maybe_thread = None
        finally:
            try:
                self._bus.shutdown()
            except Exception as ex:
                _logger.exception("%s: Bus closing error: %s", self, ex)

    @staticmethod
    def list_available_interface_names() -> typing.Iterable[str]:
        """
        Returns an empty list. TODO: provide minimally functional implementation.
        """
        return []

    def _invoke_rx_handler(self, frs: typing.List[typing.Tuple[Timestamp, Envelope]]) -> None:
        try:
            # Don't call after closure to prevent race conditions and use-after-close.
            if not self._closed and self._rx_handler is not None:
                self._rx_handler(frs)
        except Exception as exc:
            _logger.exception("%s unhandled exception in the receive handler: %s; lost frames: %s", self, exc, frs)

    def _thread_function(self, loop: asyncio.AbstractEventLoop) -> None:
        while not self._closed:
            try:
                batch = self._read_batch()
                if batch:
                    try:
                        loop.call_soon_threadsafe(self._invoke_rx_handler, batch)
                    except RuntimeError as ex:
                        _logger.debug("%s: Event loop is closed, exiting: %r", self, ex)
                        break
            except OSError as ex:
                if not self._closed:
                    _logger.exception("%s thread input/output error; stopping: %s", self, ex)
                break
            except Exception as ex:
                _logger.exception("%s thread failure: %s", self, ex)
                if not self._closed:
                    time.sleep(1)  # Is this an adequate failure management strategy?

        self._closed = True
        _logger.info("%s thread is about to exit", self)

    def _read_batch(self) -> typing.List[typing.Tuple[Timestamp, Envelope]]:
        batch: typing.List[typing.Tuple[Timestamp, Envelope]] = []
        while not self._closed:
            msg = self._bus.recv(0.0 if batch else self._MAXIMAL_TIMEOUT_SEC)
            if msg is None:
                break
            timestamp = Timestamp.now()  # TODO: use accurate timestamping
            loopback = False  # TODO: no possibility to get real loopback yet
            frame = self._parse_native_frame(msg)
            if frame is not None:
                batch.append((timestamp, Envelope(frame, loopback)))
        return batch

    @staticmethod
    def _parse_native_frame(msg: can.Message) -> typing.Optional[DataFrame]:
        if msg.is_error_frame:  # error frame, ignore silently
            _logger.debug("Error frame dropped: id_raw=%08x", msg.arbitration_id)
            return None
        frame_format = FrameFormat.EXTENDED if msg.is_extended_id else FrameFormat.BASE
        data = msg.data
        return DataFrame(frame_format, msg.arbitration_id, data)


@dataclasses.dataclass(frozen=True)
class _InterfaceParameters:
    interface_name: str
    channel_name: str


@dataclasses.dataclass(frozen=True)
class _ClassicInterfaceParameters(_InterfaceParameters):
    bitrate: int


@dataclasses.dataclass(frozen=True)
class _FDInterfaceParameters(_InterfaceParameters):
    bitrate: typing.Tuple[int, int]


def _construct_socketcan(parameters: _InterfaceParameters) -> can.ThreadSafeBus:
    if isinstance(parameters, _ClassicInterfaceParameters):
        return can.ThreadSafeBus(interface=parameters.interface_name, channel=parameters.channel_name, fd=False)
    if isinstance(parameters, _FDInterfaceParameters):
        return can.ThreadSafeBus(interface=parameters.interface_name, channel=parameters.channel_name, fd=True)
    assert False, "Internal error"


def _construct_kvaser(parameters: _InterfaceParameters) -> can.ThreadSafeBus:
    if isinstance(parameters, _ClassicInterfaceParameters):
        return can.ThreadSafeBus(
            interface=parameters.interface_name,
            channel=parameters.channel_name,
            bitrate=parameters.bitrate,
            fd=False,
        )
    if isinstance(parameters, _FDInterfaceParameters):
        return can.ThreadSafeBus(
            interface=parameters.interface_name,
            channel=parameters.channel_name,
            bitrate=parameters.bitrate[0],
            fd=True,
            data_bitrate=parameters.bitrate[1],
        )
    assert False, "Internal error"


def _construct_slcan(parameters: _InterfaceParameters) -> can.ThreadSafeBus:
    if isinstance(parameters, _ClassicInterfaceParameters):
        return can.ThreadSafeBus(
            interface=parameters.interface_name,
            channel=parameters.channel_name,
            bitrate=parameters.bitrate,
        )
    if isinstance(parameters, _FDInterfaceParameters):
        raise InvalidMediaConfigurationError(f"Interface does not support CAN FD: {parameters.interface_name}")
    assert False, "Internal error"


def _construct_pcan(parameters: _InterfaceParameters) -> can.ThreadSafeBus:
    if isinstance(parameters, _ClassicInterfaceParameters):
        return can.ThreadSafeBus(
            interface=parameters.interface_name,
            channel=parameters.channel_name,
            bitrate=parameters.bitrate,
        )
    if isinstance(parameters, _FDInterfaceParameters):
        # These magic numbers come from the settings of PCAN adapter.
        # They don't allow any direct baudrate settings, you have to set all lengths and value of the main frequency.
        # Bit lengths below are very universal and can be applied for almost every popular baudrate.
        # There is probably a better solution here, but it needs significantly more time to implement it.
        f_clock = 40000000
        nom_tseg1, nom_tseg2, nom_sjw = 3, 1, 1
        data_tseg1, data_tseg2, data_sjw = 3, 1, 1

        nom_br = int(f_clock / parameters.bitrate[0] / (nom_tseg1 + nom_tseg2 + nom_sjw))
        data_br = int(f_clock / parameters.bitrate[1] / (data_tseg1 + data_tseg2 + data_sjw))
        # TODO: validate the result and see if it is within an acceptable range

        return can.ThreadSafeBus(
            interface=parameters.interface_name,
            channel=parameters.channel_name,
            f_clock=f_clock,
            nom_brp=nom_br,
            data_brp=data_br,
            nom_tseg1=nom_tseg1,
            nom_tseg2=nom_tseg2,
            nom_sjw=nom_sjw,
            data_tseg1=data_tseg1,
            data_tseg2=data_tseg2,
            data_sjw=data_sjw,
            fd=True,
        )

    assert False, "Internal error"


def _construct_virtual(parameters: _InterfaceParameters) -> can.ThreadSafeBus:
    if isinstance(parameters, _ClassicInterfaceParameters):
        return can.ThreadSafeBus(interface=parameters.interface_name, bitrate=parameters.bitrate)
    if isinstance(parameters, _FDInterfaceParameters):
        return can.ThreadSafeBus(interface=parameters.interface_name)
    assert False, "Internal error"


def _construct_usb2can(parameters: _InterfaceParameters) -> can.ThreadSafeBus:
    if isinstance(parameters, _ClassicInterfaceParameters):
        return can.ThreadSafeBus(
            interface=parameters.interface_name,
            channel=parameters.channel_name,
            bitrate=parameters.bitrate,
        )
    if isinstance(parameters, _FDInterfaceParameters):
        raise InvalidMediaConfigurationError(f"Interface does not support CAN FD: {parameters.interface_name}")
    assert False, "Internal error"


def _construct_canalystii(parameters: _InterfaceParameters) -> can.ThreadSafeBus:
    if isinstance(parameters, _ClassicInterfaceParameters):
        return can.ThreadSafeBus(
            interface=parameters.interface_name, channel=parameters.channel_name, bitrate=parameters.bitrate
        )
    if isinstance(parameters, _FDInterfaceParameters):
        raise InvalidMediaConfigurationError(f"Interface does not support CAN FD: {parameters.interface_name}")
    assert False, "Internal error"


def _construct_any(parameters: _InterfaceParameters) -> can.ThreadSafeBus:
    raise InvalidMediaConfigurationError(f"Interface not supported yet: {parameters.interface_name}")


_CONSTRUCTORS: typing.DefaultDict[
    str, typing.Callable[[_InterfaceParameters], can.ThreadSafeBus]
] = collections.defaultdict(
    lambda: _construct_any,
    {
        "socketcan": _construct_socketcan,
        "kvaser": _construct_kvaser,
        "slcan": _construct_slcan,
        "pcan": _construct_pcan,
        "virtual": _construct_virtual,
        "usb2can": _construct_usb2can,
        "canalystii": _construct_canalystii,
    },
)



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import abc
import typing


class CRCAlgorithm(abc.ABC):
    """
    Implementations are default-constructible.
    """

    @abc.abstractmethod
    def add(self, data: typing.Union[bytes, bytearray, memoryview]) -> None:
        """
        Updates the value with the specified block of data.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def check_residue(self) -> bool:
        """
        Checks if the current state matches the algorithm-specific residue.
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def value(self) -> int:
        """
        The current CRC value, with output XOR applied, if applicable.
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def value_as_bytes(self) -> bytes:
        """
        The current CRC value serialized in the algorithm-specific byte order.
        """
        raise NotImplementedError

    @classmethod
    def new(cls, *fragments: typing.Union[bytes, bytearray, memoryview]) -> CRCAlgorithm:
        """
        A factory that creates the new instance with the value computed over the fragments.
        """
        self = cls()
        for frag in fragments:
            self.add(frag)
        return self



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import typing
import asyncio
import warnings
import dataclasses
import pyuavcan.transport
import pyuavcan.util
from ._input_session import LoopbackInputSession
from ._output_session import LoopbackOutputSession
from ._tracer import LoopbackCapture, LoopbackTracer


@dataclasses.dataclass
class LoopbackTransportStatistics(pyuavcan.transport.TransportStatistics):
    pass


class LoopbackTransport(pyuavcan.transport.Transport):
    """
    The loopback transport is intended for basic testing and API usage demonstrations.
    It works by short-circuiting input and output sessions together as if there was an underlying network.

    It is not possible to exchange data between different nodes using this transport.
    The only valid usage is sending and receiving same data on the same node.
    """

    def __init__(
        self,
        local_node_id: typing.Optional[int],
        *,
        allow_anonymous_transfers: bool = True,
        loop: typing.Optional[asyncio.AbstractEventLoop] = None,
    ):
        if loop:
            warnings.warn("The loop argument is deprecated", DeprecationWarning)
        self._local_node_id = int(local_node_id) if local_node_id is not None else None
        self._allow_anonymous_transfers = allow_anonymous_transfers
        self._input_sessions: typing.Dict[pyuavcan.transport.InputSessionSpecifier, LoopbackInputSession] = {}
        self._output_sessions: typing.Dict[pyuavcan.transport.OutputSessionSpecifier, LoopbackOutputSession] = {}
        self._capture_handlers: typing.List[pyuavcan.transport.CaptureCallback] = []
        self._spoof_result: typing.Union[bool, Exception] = True
        # Unlimited protocol capabilities by default.
        self._protocol_parameters = pyuavcan.transport.ProtocolParameters(
            transfer_id_modulo=2 ** 64,
            max_nodes=2 ** 64,
            mtu=2 ** 64 - 1,
        )

    @property
    def protocol_parameters(self) -> pyuavcan.transport.ProtocolParameters:
        return self._protocol_parameters

    @protocol_parameters.setter
    def protocol_parameters(self, value: pyuavcan.transport.ProtocolParameters) -> None:
        if isinstance(value, pyuavcan.transport.ProtocolParameters):
            self._protocol_parameters = value
        else:  # pragma: no cover
            raise ValueError(f"Unexpected value: {value}")

    @property
    def local_node_id(self) -> typing.Optional[int]:
        return self._local_node_id

    @property
    def spoof_result(self) -> typing.Union[bool, Exception]:
        """
        Test rigging. If True, :meth:`spoof` will always succeed (this is the default).
        If False, it will always time out. If :class:`Exception`, it will be raised.
        """
        return self._spoof_result

    @spoof_result.setter
    def spoof_result(self, value: typing.Union[bool, Exception]) -> None:
        self._spoof_result = value

    def close(self) -> None:
        sessions = (*self._input_sessions.values(), *self._output_sessions.values())
        self._input_sessions.clear()
        self._output_sessions.clear()
        for s in sessions:
            s.close()
        self.spoof_result = pyuavcan.transport.ResourceClosedError(f"The transport is closed: {self}")

    def get_input_session(
        self, specifier: pyuavcan.transport.InputSessionSpecifier, payload_metadata: pyuavcan.transport.PayloadMetadata
    ) -> LoopbackInputSession:
        def do_close() -> None:
            try:
                del self._input_sessions[specifier]
            except LookupError:
                pass

        try:
            sess = self._input_sessions[specifier]
        except KeyError:
            sess = LoopbackInputSession(specifier=specifier, payload_metadata=payload_metadata, closer=do_close)
            self._input_sessions[specifier] = sess
        return sess

    def get_output_session(
        self, specifier: pyuavcan.transport.OutputSessionSpecifier, payload_metadata: pyuavcan.transport.PayloadMetadata
    ) -> LoopbackOutputSession:
        def do_close() -> None:
            try:
                del self._output_sessions[specifier]
            except LookupError:
                pass

        async def do_route(tr: pyuavcan.transport.Transfer, monotonic_deadline: float) -> bool:
            _ = monotonic_deadline  # Unused, all operations always successful and instantaneous.
            if specifier.remote_node_id in {self.local_node_id, None}:  # Otherwise drop the transfer.
                tr_from = pyuavcan.transport.TransferFrom(
                    timestamp=tr.timestamp,
                    priority=tr.priority,
                    transfer_id=tr.transfer_id % self.protocol_parameters.transfer_id_modulo,
                    fragmented_payload=list(tr.fragmented_payload),
                    source_node_id=self.local_node_id,
                )
                del tr
                pyuavcan.util.broadcast(self._capture_handlers)(
                    LoopbackCapture(
                        tr_from.timestamp,
                        pyuavcan.transport.AlienTransfer(
                            pyuavcan.transport.AlienTransferMetadata(
                                tr_from.priority,
                                tr_from.transfer_id,
                                pyuavcan.transport.AlienSessionSpecifier(
                                    self.local_node_id, specifier.remote_node_id, specifier.data_specifier
                                ),
                            ),
                            list(tr_from.fragmented_payload),
                        ),
                    )
                )
                # Multicast to both: selective and promiscuous.
                for remote_node_id in {self.local_node_id, None}:  # pylint: disable=use-sequence-for-iteration
                    try:
                        destination_session = self._input_sessions[
                            pyuavcan.transport.InputSessionSpecifier(specifier.data_specifier, remote_node_id)
                        ]
                    except LookupError:
                        pass
                    else:
                        await destination_session.push(tr_from)
            return True

        try:
            sess = self._output_sessions[specifier]
        except KeyError:
            if self.local_node_id is None and not self._allow_anonymous_transfers:
                raise pyuavcan.transport.OperationNotDefinedForAnonymousNodeError(
                    f"Anonymous transfers are not enabled for {self}"
                ) from None
            sess = LoopbackOutputSession(
                specifier=specifier, payload_metadata=payload_metadata, closer=do_close, router=do_route
            )
            self._output_sessions[specifier] = sess
        return sess

    def sample_statistics(self) -> LoopbackTransportStatistics:
        return LoopbackTransportStatistics()

    @property
    def input_sessions(self) -> typing.Sequence[LoopbackInputSession]:
        return list(self._input_sessions.values())

    @property
    def output_sessions(self) -> typing.Sequence[LoopbackOutputSession]:
        return list(self._output_sessions.values())

    def begin_capture(self, handler: pyuavcan.transport.CaptureCallback) -> None:
        self._capture_handlers.append(handler)

    @property
    def capture_active(self) -> bool:
        return len(self._capture_handlers) > 0

    @staticmethod
    def make_tracer() -> LoopbackTracer:
        """
        See :class:`LoopbackTracer`.
        """
        return LoopbackTracer()

    async def spoof(self, transfer: pyuavcan.transport.AlienTransfer, monotonic_deadline: float) -> bool:
        """
        Spoofed transfers can be observed using :meth:`begin_capture`. Also see :attr:`spoof_result`.
        """
        if isinstance(self._spoof_result, Exception):
            raise self._spoof_result
        if self._spoof_result:
            pyuavcan.util.broadcast(self._capture_handlers)(
                LoopbackCapture(pyuavcan.transport.Timestamp.now(), transfer)
            )
        else:
            await asyncio.sleep(monotonic_deadline - asyncio.get_running_loop().time())
        return self._spoof_result

    @property
    def capture_handlers(self) -> typing.Sequence[pyuavcan.transport.CaptureCallback]:
        return self._capture_handlers[:]

    def _get_repr_fields(self) -> typing.Tuple[typing.List[typing.Any], typing.Dict[str, typing.Any]]:
        return [], {
            "local_node_id": self.local_node_id,
            "allow_anonymous_transfers": self._allow_anonymous_transfers,
        }



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from ._base import Port as Port
from ._base import Closable as Closable
from ._base import MessagePort as MessagePort
from ._base import ServicePort as ServicePort
from ._base import DEFAULT_PRIORITY as DEFAULT_PRIORITY
from ._base import DEFAULT_SERVICE_REQUEST_TIMEOUT as DEFAULT_SERVICE_REQUEST_TIMEOUT
from ._base import OutgoingTransferIDCounter as OutgoingTransferIDCounter
from ._base import PortFinalizer as PortFinalizer

from ._publisher import Publisher as Publisher
from ._publisher import PublisherImpl as PublisherImpl

from ._subscriber import Subscriber as Subscriber
from ._subscriber import SubscriberImpl as SubscriberImpl
from ._subscriber import SubscriberStatistics as SubscriberStatistics

from ._client import Client as Client
from ._client import ClientImpl as ClientImpl
from ._client import ClientStatistics as ClientStatistics

from ._server import Server as Server
from ._server import ServerStatistics as ServerStatistics
from ._server import ServiceRequestMetadata as ServiceRequestMetadata
from ._server import ServiceRequestHandler as ServiceRequestHandler

from ._error import PortClosedError as PortClosedError
from ._error import RequestTransferIDVariabilityExhaustedError as RequestTransferIDVariabilityExhaustedError



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from ._base import SessionFinalizer as SessionFinalizer

from ._input import CANInputSession as CANInputSession
from ._input import CANInputSessionStatistics as CANInputSessionStatistics

from ._output import CANOutputSession as CANOutputSession
from ._output import BroadcastCANOutputSession as BroadcastCANOutputSession
from ._output import UnicastCANOutputSession as UnicastCANOutputSession
from ._output import SendTransaction as SendTransaction

from ._transfer_reassembler import TransferReassemblyErrorID as TransferReassemblyErrorID
from ._transfer_reassembler import TransferReassembler as TransferReassembler

from ._transfer_sender import serialize_transfer as serialize_transfer
